/**
 * asusgeneral.c
 *
 * Implements the ASUSGENERAL bit/byte/word controller chip/sensors
 * Includes the i2c chip object as well as the sensor object
 * 
 */

#include <malloc.h>

#include <fcntl.h>

#include <pp/selector.h>

#include <pp/base.h>
#include <pp/bmc/debug.h>
#include <pp/bmc/ipmi_sdr.h>
#include <pp/bmc/topo_factory.h>
#include <pp/bmc/tp_i2c_chip.h>
#include <pp/bmc/tp_scan_sensdev.h>
#include <pp/bmc/tp_pwm_act.h>
#include <pp/bmc/tp_gpio_dev.h>

#include "drivers/asusgeneral.h"

#include <pp/bmc/bmc_dev_oem_asus.h>

/*
 * The chip, may contain global initialization code
 ****************************************************/

typedef struct asusgeneral_chip_s {
    pp_tp_i2c_chip_t base;
	unsigned char init_trx_mode;
	    // trx_mode (bit 1,0)
	    // 0 : byte
	    // 1 : byte data
	    // 2 : word data
	    // 3 : byte data twice (register offset + 1)
	    // others : reserved
	unsigned char init_slaveadd;
	unsigned char init_reg;
	unsigned short init_andmask;
	unsigned short init_or_value;
} asusgeneral_chip_t;

static void asusgeneral_dtor(pp_tp_obj_t* o);
static void asusgeneral_power_on_init(pp_tp_i2c_chip_t* this);
static int asusgeneral_power_on_configure(const int item_id UNUSED, void* ctx);

pp_tp_obj_t* pp_sensor_asusgeneral_ctor(const char* id, vector_t* args) {
    pp_strstream_t err = PP_STRSTREAM_INITIALIZER;
    asusgeneral_chip_t* o = NULL;
    pp_tp_i2c_comdev_t* i2cdev;
    int i2caddr;
    pp_tp_cond_t* icond;
	unsigned char init_trx_mode = 0;
	unsigned char init_slaveadd = 0;
	unsigned char init_reg = 0;
	unsigned short init_andmask = 0;
	unsigned short init_or_value = 0;
    
    if (pp_tp_arg_scanf(args, 0, &err, "o<i>do<sc>", &i2cdev, &i2caddr, &icond) != 3) {
		pp_bmc_log_perror("%s(): '%s' failed: %s at stage %d", ___F, id, pp_strstream_buf(&err), 1);
    } else if (pp_tp_arg_scanf(args, 3, &err, "|d<c>d<c>d<c>d<s>d<s>", &init_trx_mode, &init_slaveadd, &init_reg, &init_andmask, &init_or_value) != 5) {
		pp_bmc_log_perror("%s(): '%s' failed: %s at stage %d", ___F, id, pp_strstream_buf(&err), 2);
    } else {
		o = malloc(sizeof(asusgeneral_chip_t));
		o->init_trx_mode = init_trx_mode;
		o->init_slaveadd = init_slaveadd;
		o->init_reg = init_reg;
		o->init_andmask = init_andmask;
		o->init_or_value = init_or_value;
		pp_tp_i2c_chip_init(&o->base, PP_TP_I2C_CHIP, id, asusgeneral_dtor,
			    ASUSGENERAL_MODEL_STRING, i2cdev, i2caddr, icond,
			    asusgeneral_power_on_init, NULL);
    };
    
    pp_strstream_free(&err);
    return (pp_tp_obj_t*)o;
}

static void asusgeneral_dtor(pp_tp_obj_t* o) {
    pp_tp_i2c_chip_t* this = (pp_tp_i2c_chip_t*)o;
    assert(this != NULL);
    pp_tp_i2c_chip_cleanup(this);
    free(this);
}

static void asusgeneral_power_on_init(pp_tp_i2c_chip_t* this) {
    pp_bmc_log_notice("[ASUSGENERAL] power_on_init");
    pp_select_add_to(0, 0, asusgeneral_power_on_configure, (void*)this);
}

static int asusgeneral_power_on_configure(const int item_id UNUSED, void* ctx) {
    int ret = PP_SUC;
    asusgeneral_chip_t* this = (asusgeneral_chip_t*)ctx;
    unsigned char i2cadd_backup = this->base.i2caddr;
    unsigned char data_byte;
    unsigned short data_word;
    static char retry_cnt = 0;
    
    if (this->init_slaveadd != 0) {
		this->base.i2caddr = (this->init_slaveadd) >> 1;
	    if (PP_SUC == (ret = pp_tp_i2c_chip_pre_com(&this->base))) {
	    	switch (this->init_trx_mode) {
	    		case 0x00 :	// byte
	    			ret = pp_tp_i2c_chip_rx_byte(&this->base, &data_byte);
	    			if (ret == PP_SUC) {
	    				data_byte &= (unsigned char)(this->init_andmask);
	    				data_byte |= (unsigned char)(this->init_or_value);
	    				ret = pp_tp_i2c_chip_tx_byte(&this->base, data_byte);
	    			};
	    			break;
	    		case 0x01 :	// byte date
	    			ret = pp_tp_i2c_chip_rx_byte_data(&this->base, this->init_reg, &data_byte);
	    			if (ret == PP_SUC) {
	    				data_byte &= (unsigned char)(this->init_andmask);
	    				data_byte |= (unsigned char)(this->init_or_value);
	    				ret = pp_tp_i2c_chip_tx_byte_data(&this->base, this->init_reg, data_byte);
	    			};
	    			break;
	    		case 0x02 :	// word data
	    			ret = pp_tp_i2c_chip_rx_word_data(&this->base, this->init_reg, &data_word);
	    			if (ret == PP_SUC) {
	    				data_word &= this->init_andmask;
	    				data_word |= this->init_or_value;
	    				ret = pp_tp_i2c_chip_tx_word_data(&this->base, this->init_reg, data_word);
	    			};
	    			break;
	    		case 0x03 :	// byte data twice (register offset + 1)
	    			ret = pp_tp_i2c_chip_rx_byte_data(&this->base, this->init_reg, &data_byte);
	    			if (ret == PP_SUC) {
	    				data_word = (unsigned short)data_byte;
		    			ret = pp_tp_i2c_chip_rx_byte_data(&this->base, (this->init_reg)+1, &data_byte);
		    			if (ret == PP_SUC) {
		    				data_word |= ((unsigned short)data_byte) << 8;
		    				data_word &= this->init_andmask;
		    				data_word |= this->init_or_value;
		    				data_byte = (unsigned char)(data_word & 0x00FF);
		    				ret = pp_tp_i2c_chip_tx_byte_data(&this->base, this->init_reg, data_byte);
			    			if (ret == PP_SUC) {
			    				data_byte = (unsigned char)((data_word & 0xFF00) >> 8);
			    				ret = pp_tp_i2c_chip_tx_byte_data(&this->base, (this->init_reg)+1, data_byte);
			    			};
			    		};
	    			};
	    			break;
	    		default:
				    pp_bmc_log_error("[ASUSGENERAL] Unknown SMBus protocol type init process");
	    			break;
	    	};
	        pp_tp_i2c_chip_post_com(&this->base);
	    };
		this->base.i2caddr = i2cadd_backup;
	};
    
    if (ret != PP_SUC) {
    	if (retry_cnt++ < 3) {
        	pp_select_add_to(1000, 0, asusgeneral_power_on_configure, (void*)this);
	    	pp_bmc_log_notice("[ASUSGENERAL] init failed, retrying in 1 second, retry count #%d", retry_cnt);
	    } else {
		    pp_bmc_log_error("[ASUSGENERAL] init process failed");
		};
    } else {
	    pp_bmc_log_notice("[ASUSGENERAL] init process success");
    };

    return PP_SUC;
}



/*
 * sensors
 ************************/
typedef struct check_s {
	union {
        struct {
            unsigned char reg;			// unused
            unsigned short andmask;
            unsigned char ok_value;
            unsigned short pass_raw_return;
            unsigned short raw_return;
            unsigned short or_value;	//for pre_enable usage
        } byte;
        struct {
            unsigned char reg;
            unsigned short andmask;
            BITFIELD5(unsigned char, check_mode : 3,checked_return_type : 1,
            		divided_byte_return : 1, numerator_denominator : 1, reserved : 2);
            //	bit	2,1,0 (checking : <, =, >)
            //			000 : don't check (raw/divided byte return)
            //			001 : >		check
            //			010 : ==	check
            //			011 : >=	check
            //			100 : <		check
            //			101 : <>	check (range checking)
            //			110 : <=	check
            //			111 : !=	check
            //	bit	3 (for checked return type usage only)
            //			0 : raw byte return	(0 for checked fail and one byte return for checked pass)
            //			1 : discrete return (0: OK, 1:Fail)
			//	bit	4 (divided-byte return)
            //			0 : raw value divided disable
            //			1 : raw value divided enable
			//	bit	5 (divisor value as numerator or denominator)
            //			0 : divisor value as denominator
            //			1 : divisor value as numerator
            unsigned short divisor_value;
            unsigned short ok_eqhigh_limit;	//equal or high limit raw value
            unsigned short ok_enlow_limit;	//pre_enable usage or low limit raw value
        } byte_data;
        struct {
            unsigned char reg;
            unsigned short andmask;
            BITFIELD6(unsigned char, check_mode : 3,checked_return_type : 1,
            		divided_byte_return : 1, numerator_denominator : 1, hilo_byte_return : 1, reserved : 1);
            //	bit	2,1,0 (checking : <, =, >)
            //			000 : don't check (raw_high/raw_low/divided byte return)
            //			001 : >		check
            //			010 : ==	check
            //			011 : >=	check
            //			100 : <		check
            //			101 : <>	check (range checking)
            //			110 : <=	check
            //			111 : !=	check
            //	bit	3 (for checked return type usage only)
            //			0 : byte return	(0 for checked fail and one byte return for checked pass)
            //			1 : discrete return (0: OK, 1:Fail)
			//	bit	4 (divided-byte return), only when bit3=0
            //			0 : raw value divided disable
            //			1 : raw value divided enable
			//	bit	5 (divisor value as numerator or denominator), only when bit3=0 & bit4=1
            //			0 : divisor value as denominator
            //			1 : divisor value as numerator
			//	bit	6 (high/low byte raw return), only when bit3=0 & bit4=0
            //			0 : low byte raw return
            //			1 : high byte raw return
            unsigned short divisor_value;
            unsigned short ok_eqhigh_limit;	//equal or high limit raw value
            unsigned short ok_enlow_limit;	//pre_enable usage or low limit raw value
        } word_data;
	};
} check_t;

typedef struct asusgeneral_sens_cfg_s {
    unsigned char sens_type;
    unsigned char reading_type;
    unsigned char unit_type;
    unsigned char analog_type;
    /* sdr configuration */
    unsigned short sdr_mtol;
    unsigned short sdr_bacc;
    char sdr_bexp;
    char sdr_rexp;
    /* configuration */
    struct {	//pre_enable : write the certain register before update sensor reading value
	    BITFIELD5(unsigned char, trx_mode_1 : 2,trx_mode_2 : 2, pre_en_trx_mode : 2, trx_2_en : 1, pre_trx_en : 1);
	    // trx_mode (bit 1,0)
	    // 0 : byte
	    // 1 : byte data
	    // 2 : word data
	    // 3 : byte data twice (register offset + 1)
	    unsigned char slave_add;
        check_t check_b[2];
        check_t pre_en;
    } config;
} asusgeneral_sens_cfg_t;

typedef struct asusgeneral_sens_s {
    pp_tp_scan_sensdev_t base;
    pp_tp_i2c_chip_t* chip;
//    asusgeneral_sens_cfg_t* cfg;
    asusgeneral_sens_cfg_t cfg;
} asusgeneral_sens_t;

static void asusgeneral_sens_destroy(pp_tp_obj_t* o);
static ipmi_sdr_header_t* asusgeneral_sens_default_sdr(pp_tp_sensdev_t* s);
static void asusgeneral_update_reading(pp_sensor_scannable_t*);

asusgeneral_sens_cfg_t asusgeneral_sens_cfg[1];

//static void printcheck(asusgeneral_sens_cfg_t *cfg)
//{
//	unsigned char idx;
//	
//	for (idx = 0; idx < sizeof(asusgeneral_sens_cfg_t); idx++) {
//		pp_bmc_log_notice("asusgeneral_sens_cfg[%2.2d]=0x%2.2X", idx, *(((unsigned char *)cfg)+idx) );
//	};
//	pp_bmc_log_info("cfg->config.trx_mode_1       = 0x%2.2X", cfg->config.trx_mode_1);
//	pp_bmc_log_info("cfg->config.trx_mode_2       = 0x%2.2X", cfg->config.trx_mode_2);
//	pp_bmc_log_info("cfg->config.pre_en_trx_mode  = 0x%2.2X", cfg->config.pre_en_trx_mode);
//	pp_bmc_log_info("cfg->config.trx_2_en         = 0x%2.2X", cfg->config.trx_2_en);
//	pp_bmc_log_info("cfg->config.pre_trx_en       = 0x%2.2X", cfg->config.pre_trx_en);
//}

pp_tp_obj_t* pp_sensor_asusgeneral_sens_ctor(const char* id, vector_t* args)
{
    const char* errmsg = "%s(): '%s' failed: %s at stage %d";
    asusgeneral_sens_t* this = NULL;
    pp_strstream_t err = PP_STRSTREAM_INITIALIZER;
    pp_tp_i2c_chip_t* chip;
    pp_tp_cond_t* scan_cond;
	asusgeneral_sens_cfg_t* cfg = &asusgeneral_sens_cfg[0];

	memset(cfg, 0, sizeof(asusgeneral_sens_cfg_t));
	
    if (pp_tp_arg_scanf(args, 0, &err, "o<h>o<sc>", &chip, &scan_cond) != 2) {
		pp_bmc_log_perror(errmsg, ___F, id, pp_strstream_buf(&err), 1);
	} else if (pp_tp_arg_scanf(args, 2, &err, "d<c>d<c>d<c>d<c>d<s>d<s>d<c>d<c>d<c>d<c>",
			&(cfg->sens_type), &(cfg->reading_type), &(cfg->unit_type), &(cfg->analog_type),
			&(cfg->sdr_mtol), &(cfg->sdr_bacc), &(cfg->sdr_bexp), &(cfg->sdr_rexp),
			&(cfg->config), &(cfg->config.slave_add)) != 10) {
		pp_bmc_log_perror(errmsg, ___F, id, pp_strstream_buf(&err), 2);
	} else if (pp_tp_arg_scanf(args, 12, &err, "d<c>d<s>d<c>d<s>d<s>d<s>|d<c>d<s>d<c>d<s>d<s>d<s>d<c>d<s>d<c>d<s>d<s>d<s>",
			&(cfg->config.check_b[0].byte.reg), &(cfg->config.check_b[0].byte.andmask), &(cfg->config.check_b[0].byte.ok_value), &(cfg->config.check_b[0].byte.pass_raw_return), &(cfg->config.check_b[0].byte.raw_return), &(cfg->config.check_b[0].byte.or_value),
			&(cfg->config.check_b[1].byte.reg), &(cfg->config.check_b[1].byte.andmask), &(cfg->config.check_b[1].byte.ok_value), &(cfg->config.check_b[1].byte.pass_raw_return), &(cfg->config.check_b[1].byte.raw_return), &(cfg->config.check_b[1].byte.or_value),
			&(cfg->config.pre_en.byte.reg), &(cfg->config.pre_en.byte.andmask), &(cfg->config.pre_en.byte.ok_value), &(cfg->config.pre_en.byte.pass_raw_return), &(cfg->config.pre_en.byte.raw_return), &(cfg->config.pre_en.byte.or_value)) != 18) {
		pp_bmc_log_perror(errmsg, ___F, id, pp_strstream_buf(&err), 3);
    } else if (strcmp(chip->model, ASUSGENERAL_MODEL_STRING)) {
		errno = EINVAL;
		pp_bmc_log_perror(errmsg, ___F, id, "incorrect chip model");
    } else {
		this = malloc(sizeof(asusgeneral_sens_t));
		pp_tp_scan_sensdev_init(&this->base, PP_TP_SCAN_SENS_DEV, id,
					(void(*)(pp_tp_obj_t*))asusgeneral_sens_destroy,
					asusgeneral_sens_default_sdr, asusgeneral_update_reading,
					scan_cond);
		this->chip = pp_tp_i2c_chip_duplicate(chip);
//		this->cfg = cfg;
		memcpy(&(this->cfg), cfg, sizeof(asusgeneral_sens_cfg_t));
		
		/* register for scanning  */
		pp_tp_scan_sensdev_register(&this->base, 0);
    };
    
    pp_strstream_free(&err);
//	printcheck(cfg);
    return (pp_tp_obj_t*)this;
}

static void asusgeneral_sens_destroy(pp_tp_obj_t* o) {
    asusgeneral_sens_t* this = (asusgeneral_sens_t*)o;
    assert(this != NULL);
    pp_tp_scan_sensdev_unregister(&this->base);
    pp_tp_i2c_chip_release(this->chip);
    pp_tp_scan_sensdev_cleanup(&this->base);
    free(this);
}

static ipmi_sdr_header_t* asusgeneral_sens_default_sdr(pp_tp_sensdev_t* s) {
    asusgeneral_sens_t* this = (asusgeneral_sens_t*)s;
//    asusgeneral_sens_cfg_t* cfg = this->cfg;
    asusgeneral_sens_cfg_t* cfg = &(this->cfg);
    ipmi_sdr_full_sensor_t* sdr;

    sdr = ipmi_sdr_full_part_create(cfg->sens_type, cfg->reading_type, cfg->unit_type, cfg->analog_type,
							    cfg->sdr_mtol, cfg->sdr_bacc, cfg->sdr_bexp, cfg->sdr_rexp,
                               ipmi_sdr_max_by_form(cfg->analog_type),
                               ipmi_sdr_min_by_form(cfg->analog_type));
/*
ipmi_sdr_full_create(unsigned char sens_type,
			unsigned char reading_type,
			unsigned char base_unit,
			unsigned char analog_format,
			unsigned short mtol,
			unsigned char bacc,
			char bexp,
			char rexp,
			unsigned char sens_max,
			unsigned char sens_min)
*/
    return (ipmi_sdr_header_t*)sdr;
}

static void asusgeneral_update_reading(pp_sensor_scannable_t* s) {
    asusgeneral_sens_t* this = PP_TP_SCANNABLE_2_OBJ_CAST(s, asusgeneral_sens_t);
    pp_tp_i2c_chip_t* chip = this->chip;
//    asusgeneral_sens_cfg_t* cfg = this->cfg;
    asusgeneral_sens_cfg_t* cfg = &(this->cfg);
    int ret = PP_ERR;
    unsigned char i2cadd_backup = chip->i2caddr;
    unsigned char data_byte, raw_data_byte = 0, value = 0, check_pass_flag = 0;
    unsigned short data_word, raw_data_word = 0;
	unsigned char pre_enable_ok = 0, check_check_ok = 0, divisor_zero = 0;
	unsigned char check_index, check_go;
    
    if (pp_tp_i2c_chip_is_initialized(chip) && pp_tp_scan_sensdev_do_scan(&this->base)) {
		chip->i2caddr = (cfg->config.slave_add) >> 1;
        if (PP_SUC == (ret = pp_tp_i2c_chip_pre_com(chip))) {
// pre_enable
			if (cfg->config.pre_trx_en == 1) {
		    	switch (cfg->config.pre_en_trx_mode) {
		    		case 0x00 :	// byte
		    			ret = pp_tp_i2c_chip_rx_byte(chip, &data_byte);
		    			if (ret == PP_SUC) {
		    				data_byte &= (unsigned char)(cfg->config.pre_en.byte.andmask);
		    				data_byte |= (unsigned char)(cfg->config.pre_en.byte.or_value);
		    				ret = pp_tp_i2c_chip_tx_byte(chip, data_byte);
		    			};
		    			break;
		    		case 0x01 :	// byte date
		    			ret = pp_tp_i2c_chip_rx_byte_data(chip, cfg->config.pre_en.byte_data.reg, &data_byte);
		    			if (ret == PP_SUC) {
		    				data_byte &= (unsigned char)(cfg->config.pre_en.byte_data.andmask);
		    				data_byte |= (unsigned char)(cfg->config.pre_en.byte_data.ok_enlow_limit);
		    				ret = pp_tp_i2c_chip_tx_byte_data(chip, cfg->config.pre_en.byte_data.reg, data_byte);
		    			};
		    			break;
		    		case 0x02 :	// word data
		    			ret = pp_tp_i2c_chip_rx_word_data(chip, cfg->config.pre_en.word_data.reg, &data_word);
		    			if (ret == PP_SUC) {
		    				data_word &= cfg->config.pre_en.word_data.andmask;
		    				data_word |= cfg->config.pre_en.word_data.ok_enlow_limit;
		    				ret = pp_tp_i2c_chip_tx_word_data(chip, cfg->config.pre_en.word_data.reg, data_word);
		    			};
		    			break;
		    		case 0x03 :	// byte data twice (register offset + 1)
		    			ret = pp_tp_i2c_chip_rx_byte_data(chip, cfg->config.pre_en.word_data.reg, &data_byte);
		    			if (ret == PP_SUC) {
		    				data_word = (unsigned short)data_byte;
			    			ret = pp_tp_i2c_chip_rx_byte_data(chip, (cfg->config.pre_en.word_data.reg)+1, &data_byte);
			    			if (ret == PP_SUC) {
			    				data_word |= ((unsigned short)data_byte) << 8;
			    				data_word &= cfg->config.pre_en.word_data.andmask;
			    				data_word |= cfg->config.pre_en.word_data.ok_enlow_limit;
			    				data_byte = (unsigned char)(data_word & 0x00FF);
			    				ret = pp_tp_i2c_chip_tx_byte_data(chip, cfg->config.pre_en.word_data.reg, data_byte);
				    			if (ret == PP_SUC) {
				    				data_byte = (unsigned char)((data_word & 0xFF00) >> 8);
				    				ret = pp_tp_i2c_chip_tx_byte_data(chip, (cfg->config.pre_en.word_data.reg)+1, data_byte);
				    			};
				    		};
		    			};
		    			break;
		    	};
		    };
		    if (ret != PP_SUC) {
			    pp_bmc_log_error("[ASUSGENERAL] sensor reading update process --> pre_enable fail");
		    } else {
		    	pre_enable_ok = 1;
		    };
// check_1, check_2
			for (check_index = 0; check_index < 2; check_index++) {
				check_go = 0;
				if (check_index == 0) {
					if (pre_enable_ok == 1)
						check_go = 1;
				} else if (check_index == 1) {
					if (pre_enable_ok == 1 && check_check_ok == 1 && cfg->config.trx_2_en == 1 && divisor_zero == 0)
						check_go =1;
				};
				if (check_go == 0) continue;
		    	switch ((check_index == 0) ? cfg->config.trx_mode_1 : cfg->config.trx_mode_2){
		    		case 0x00 :	// byte
		    			ret = pp_tp_i2c_chip_rx_byte(chip, &data_byte);
		    			if (ret == PP_SUC) {
			    			raw_data_byte = data_byte;
		    				if (cfg->config.check_b[check_index].byte.raw_return == 1) {
		    					value = raw_data_byte;
		    				} else {
			    				data_byte &= (unsigned char)(cfg->config.check_b[check_index].byte.andmask);
			    				check_pass_flag = 0;
			    				if (cfg->config.check_b[check_index].byte.pass_raw_return == 1) {
			    					if (data_byte == cfg->config.check_b[check_index].byte.ok_value) {
		    							value = raw_data_byte;
					    				check_pass_flag = 1;
		    						} else {
		    							value = 0;
		    						};
		    					} else {
			    					if (data_byte == cfg->config.check_b[check_index].byte.ok_value) {
		    							value = 0;
					    				check_pass_flag = 1;
		    						} else {
		    							value = 1;
		    						};
		    					};
	   							check_check_ok = check_pass_flag;
		    				};
		    			};
		    			break;
		    		case 0x01 :	// byte date
		    			ret = pp_tp_i2c_chip_rx_byte_data(chip, cfg->config.check_b[check_index].byte_data.reg, &data_byte);
		    			if (ret == PP_SUC) {
			    			raw_data_byte = data_byte;
		    				if (cfg->config.check_b[check_index].byte_data.check_mode == 0) {
		    					if (cfg->config.check_b[check_index].byte_data.divided_byte_return == 0) {
		    						value = raw_data_byte;
		    					} else if (cfg->config.check_b[check_index].byte_data.numerator_denominator == 0) {
		    						if (cfg->config.check_b[check_index].byte_data.divisor_value == 0) {
		    							pp_bmc_log_error("[ASUSGENERAL] sensor reading update process --> divisor = 0, check_%d, byte_data, unchecked, divisor_value", check_index);
		    							divisor_zero = 1;
		    						} else {
		    							value = (unsigned char)( (((unsigned short)raw_data_byte) / cfg->config.check_b[check_index].byte_data.divisor_value) & 0x00FF );
		    						};
		    					} else {
		    						if (raw_data_byte == 0) {
		    							pp_bmc_log_error("[ASUSGENERAL] sensor reading update process --> divisor = 0, check_%d, byte_data, unchecked, raw_data_byte", check_index);
		    							divisor_zero = 1;
		    						} else {
			    						value = (unsigned char)( (cfg->config.check_b[check_index].byte_data.divisor_value / ((unsigned short)raw_data_byte)) & 0x00FF );
		    						};
		    					};
		    				} else {
			    				data_byte &= (unsigned char)(cfg->config.check_b[check_index].byte_data.andmask);
			    				check_pass_flag = 0;
		    					switch (cfg->config.check_b[check_index].byte_data.check_mode) {
		    						case 0x01 :	// X > enlow check
		    							if (data_byte > cfg->config.check_b[check_index].byte_data.ok_enlow_limit)
		    								check_pass_flag = 1;
		    							break;
		    						case 0x02 :	// X == eqhigh check
		    							if (data_byte == cfg->config.check_b[check_index].byte_data.ok_eqhigh_limit)
		    								check_pass_flag = 1;
		    							break;
		    						case 0x03 :	// X >= enlow check
		    							if (data_byte >= cfg->config.check_b[check_index].byte_data.ok_enlow_limit)
		    								check_pass_flag = 1;
		    							break;
		    						case 0x04 :	// X < eqhigh check
		    							if (data_byte < cfg->config.check_b[check_index].byte_data.ok_eqhigh_limit)
		    								check_pass_flag = 1;
		    							break;
		    						case 0x05 :	// eqhigh > X > enlow check
		    							if ((data_byte > cfg->config.check_b[check_index].byte_data.ok_enlow_limit) &&
		    								(data_byte < cfg->config.check_b[check_index].byte_data.ok_eqhigh_limit))
		    								check_pass_flag = 1;
		    							break;
		    						case 0x06 :	// X <= eqhigh check
		    							if (data_byte <= cfg->config.check_b[check_index].byte_data.ok_eqhigh_limit)
		    								check_pass_flag = 1;
		    							break;
		    						case 0x07 :	// X != eqhigh check
		    							if (data_byte != cfg->config.check_b[check_index].byte_data.ok_eqhigh_limit)
		    								check_pass_flag = 1;
		    							break;
		    					};
		    					check_check_ok = check_pass_flag;
		    					if (cfg->config.check_b[check_index].byte_data.checked_return_type == 0) {
		    						if (check_pass_flag == 0) {
		    							value = 0;
			    					} else if (cfg->config.check_b[check_index].byte_data.divided_byte_return == 0) {
			    						value = raw_data_byte;
			    					} else if (cfg->config.check_b[check_index].byte_data.numerator_denominator == 0) {
			    						if (cfg->config.check_b[check_index].byte_data.divisor_value == 0) {
			    							pp_bmc_log_error("[ASUSGENERAL] sensor reading update process --> divisor = 0, check_%d, byte_data, checked, divisor_value", check_index);
			    							divisor_zero = 1;
			    						} else {
			    							value = (unsigned char)( (((unsigned short)raw_data_byte) / cfg->config.check_b[check_index].byte_data.divisor_value) & 0x00FF );
			    						};
			    					} else {
			    						if (raw_data_byte == 0) {
			    							pp_bmc_log_error("[ASUSGENERAL] sensor reading update process --> divisor = 0, check_%d, byte_data, checked, raw_data_byte", check_index);
			    							divisor_zero = 1;
			    						} else {
				    						value = (unsigned char)( (cfg->config.check_b[check_index].byte_data.divisor_value / ((unsigned short)raw_data_byte)) & 0x00FF );
			    						};
			    					};
		    					} else {
		    						value = (check_pass_flag == 1) ? 0 : 1;
		    					};
			    			};
		    			};
		    			break;
		    		case 0x02 :	// word data
		    		case 0x03 :	// byte data twice (register offset + 1)
		    			if (((check_index == 0) ? cfg->config.trx_mode_1 : cfg->config.trx_mode_2) == 2) {
		    				ret = pp_tp_i2c_chip_rx_word_data(chip, cfg->config.check_b[check_index].word_data.reg, &data_word);
		    			} else {
			    			ret = pp_tp_i2c_chip_rx_byte_data(chip, cfg->config.check_b[check_index].word_data.reg, &data_byte);
			    			if (ret == PP_SUC) {
			    				data_word = (unsigned short)data_byte;
				    			ret = pp_tp_i2c_chip_rx_byte_data(chip, (cfg->config.check_b[check_index].word_data.reg)+1, &data_byte);
				    			if (ret == PP_SUC) {
				    				data_word |= ((unsigned short)data_byte) << 8;
				    			};
				    		};
				    	};
		    			if (ret == PP_SUC) {
			    			raw_data_word = data_word;
		    				if (cfg->config.check_b[check_index].word_data.check_mode == 0) {
		    					if (cfg->config.check_b[check_index].word_data.divided_byte_return == 0) {
		    						if (cfg->config.check_b[check_index].word_data.hilo_byte_return == 0) {
		    							value = (unsigned char)(raw_data_word & 0x00FF);
		    						} else {
		    							value = (unsigned char)((raw_data_word & 0xFF00) >> 8);
		    						};
		    					} else {
			    					if (cfg->config.check_b[check_index].word_data.numerator_denominator == 0) {
			    						if (cfg->config.check_b[check_index].word_data.divisor_value == 0) {
			    							pp_bmc_log_error("[ASUSGENERAL] sensor reading update process --> divisor = 0, check_%d, word data(byte data twice), unchecked, divisor_value", check_index);
			    							divisor_zero = 1;
			    						} else {
			    							value = (unsigned char)((raw_data_word / cfg->config.check_b[check_index].word_data.divisor_value) & 0x00FF);
			    						};
			    					} else {
			    						if (raw_data_word == 0) {
			    							pp_bmc_log_error("[ASUSGENERAL] sensor reading update process --> divisor = 0, check_%d, word data(byte data twice), unchecked, raw_data_word", check_index);
			    							divisor_zero = 1;
			    						} else {
				    						value = (unsigned char)((cfg->config.check_b[check_index].word_data.divisor_value / raw_data_word) & 0x00FF);
			    						};
			    					};
		    					};
		    				} else {
			    				data_word &= cfg->config.check_b[check_index].word_data.andmask;
			    				check_pass_flag = 0;
		    					switch (cfg->config.check_b[check_index].word_data.check_mode) {
		    						case 0x01 :	// X > enlow check
		    							if (data_word > cfg->config.check_b[check_index].word_data.ok_enlow_limit)
		    								check_pass_flag = 1;
		    							break;
		    						case 0x02 :	// X == eqhigh check
		    							if (data_word == cfg->config.check_b[check_index].word_data.ok_eqhigh_limit)
		    								check_pass_flag = 1;
		    							break;
		    						case 0x03 :	// X >= enlow check
		    							if (data_word >= cfg->config.check_b[check_index].word_data.ok_enlow_limit)
		    								check_pass_flag = 1;
		    							break;
		    						case 0x04 :	// X < eqhigh check
		    							if (data_word < cfg->config.check_b[check_index].word_data.ok_eqhigh_limit)
		    								check_pass_flag = 1;
		    							break;
		    						case 0x05 :	// eqhigh > X > enlow check
		    							if ((data_word > cfg->config.check_b[check_index].word_data.ok_enlow_limit) &&
		    								(data_word < cfg->config.check_b[check_index].word_data.ok_eqhigh_limit))
		    								check_pass_flag = 1;
		    							break;
		    						case 0x06 :	// X <= eqhigh check
		    							if (data_word <= cfg->config.check_b[check_index].word_data.ok_eqhigh_limit)
		    								check_pass_flag = 1;
		    							break;
		    						case 0x07 :	// X != eqhigh check
		    							if (data_word != cfg->config.check_b[check_index].word_data.ok_eqhigh_limit)
		    								check_pass_flag = 1;
		    							break;
		    					};
		    					check_check_ok = check_pass_flag;
		    					if (cfg->config.check_b[check_index].word_data.checked_return_type == 0) {
		    						if (check_pass_flag == 0) {
		    							value = 0;
			    					} else if (cfg->config.check_b[check_index].word_data.divided_byte_return == 0) {
			    						if (cfg->config.check_b[check_index].word_data.hilo_byte_return == 0) {
			    							value = (unsigned char)(raw_data_word & 0x00FF);
			    						} else {
			    							value = (unsigned char)((raw_data_word & 0xFF00) >> 8);
			    						};
			    					} else {
				    					if (cfg->config.check_b[check_index].word_data.numerator_denominator == 0) {
				    						if (cfg->config.check_b[check_index].word_data.divisor_value == 0) {
				    							pp_bmc_log_error("[ASUSGENERAL] sensor reading update process --> divisor = 0, check_%d, word data(byte data twice), checked, divisor_value", check_index);
				    							divisor_zero = 1;
				    						} else {
				    							value = (unsigned char)((raw_data_word / cfg->config.check_b[check_index].word_data.divisor_value) & 0x00FF);
				    						};
				    					} else {
				    						if (raw_data_word == 0) {
				    							pp_bmc_log_error("[ASUSGENERAL] sensor reading update process --> divisor = 0, check_%d, word data(byte data twice), checked, raw_data_word", check_index);
				    							divisor_zero = 1;
				    						} else {
					    						value = (unsigned char)((cfg->config.check_b[check_index].word_data.divisor_value / raw_data_word) & 0x00FF);
				    						};
				    					};
			    					};
		    					} else {
		    						value = (check_pass_flag == 1) ? 0 : 1;
		    					};
		    				};
		    			};
		    			break;
		    	};
//				pp_bmc_log_notice("[ASUSGENERAL] check_check_ok = 0x%2.2X", check_check_ok);
	    	};
            pp_tp_i2c_chip_post_com(chip);
        };
       	chip->i2caddr = i2cadd_backup;
       	if (ret == PP_SUC && pre_enable_ok == 1 && divisor_zero == 0) {
       		ret = value;
       	} else {
       		ret = PP_ERR;
       	};
    };

    if ( ( ret==0 && cfg->config.slave_add == 0x5e )  )
       ret = -2;
    pp_tp_scan_sensdev_update_reading(&this->base, ret);
}


//ASUS writable sensor table generation
pp_tp_obj_t* pp_asus_sensor_asuswritable_ctor(const char* id, vector_t* args) {
    pp_tp_obj_t *base = NULL;
    const char *cname = "[ASUSWritableSens]";
    pp_strstream_t err = PP_STRSTREAM_INITIALIZER;
    int fd, n;
	asus_write_sensor_info_t asus_write_sensor_new;
    unsigned char channel_number, sdr_check, protocol_type;
	static unsigned int asus_write_sensor_add_cnt = 0;

    memset(&asus_write_sensor_new, 0, sizeof(asus_write_sensor_new));
    // scan args
    if (pp_tp_arg_scanf(args, 0, &err, "d<c>d<c>d<c>d<c>d<c>d<c>d<c>d<c>",
    			&(asus_write_sensor_new.sensor_number),
    			&channel_number,
    			&(asus_write_sensor_new.slave_add),
    			&(asus_write_sensor_new.cmd_offset_byte[0]),
    			&(asus_write_sensor_new.cmd_offset_byte[1]),
    			&(asus_write_sensor_new.write_value_mask),
    			&sdr_check,
    			&protocol_type) != 8) {
        pp_bmc_log_perror("%s (%s) failed: %s ", cname, id, pp_strstream_buf(&err));
        goto finally_asus_sensor_asuswritable;
    }

//    base = malloc(sizeof(pp_tp_obj_t));
//    pp_tp_obj_init(base, PP_TP_SDR, id, pp_asus_sensor_asuswritable_dtor);

	if (asus_write_sensor_add_cnt < ASUS_WRITE_SENSOR_TABLE_MAX_NUMBER) {
		switch (channel_number) {
			case 0x04:
			case 0x05:
			case 0x06:
			case 0x07:
				asus_write_sensor_new.pca9545_sw_value = 1 << (channel_number - 4);
				break;
			default:
				asus_write_sensor_new.pca9545_sw_value = 0;
				break;
		};
		asus_write_sensor_new.slave_add >>= 1;
		asus_write_sensor_new.sdr_check = sdr_check;
		asus_write_sensor_new.protocol_type = protocol_type;
	    /* open file */
	    assert(ASUS_WRITE_SENSOR_TABLE_FILE_PATH);
	    if (0 > (fd = open(ASUS_WRITE_SENSOR_TABLE_FILE_PATH, O_WRONLY | O_CREAT | O_APPEND, S_RW_RW_RW_))) {
	        pp_bmc_log_perror("%s can't open file %s:", cname, ASUS_WRITE_SENSOR_TABLE_FILE_PATH);
	        goto finally_asus_sensor_asuswritable;
	    };
	    /* write file */
	    n = write(fd, &asus_write_sensor_new, sizeof(asus_write_sensor_new));
	    if ((n < 0) || (n != sizeof(asus_write_sensor_new))) {
	        pp_bmc_log_error("%s write file %s error", cname, ASUS_WRITE_SENSOR_TABLE_FILE_PATH);
	        close(fd);
	        goto finally_asus_sensor_asuswritable;
	    };
        close(fd);
        pp_bmc_log_info("%s Entry add to file %s, cnt = %d",
        				 cname, ASUS_WRITE_SENSOR_TABLE_FILE_PATH, ++asus_write_sensor_add_cnt);
        pp_bmc_log_notice("%s snr_number = 0x%2.2X, sdr_check = %d, protocol_type = %d",
	        			cname,
	        			asus_write_sensor_new.sensor_number,
	        			sdr_check,
	        			protocol_type);
	};

 finally_asus_sensor_asuswritable:
    pp_strstream_free(&err);
    return base;
}

//static void pp_asus_sensor_asuswritable(pp_tp_obj_t* o) {    
//    free (o);
//}


//ASUS Mother Board Support Model List generation
pp_tp_obj_t* pp_asus_mb_support_model_ctor(const char* id, vector_t* args) {
    pp_tp_obj_t *base = NULL;
    const char *cname = "[ASUSMBSupMdl]";
    pp_strstream_t err = PP_STRSTREAM_INITIALIZER;
    int fd, n;
    unsigned char *models[ASUS_MB_SUPPORT_MODEL_MAX_NUMBER];
    unsigned char mdl_string[ASUS_MB_SUPPORT_MODEL_MAX_LENGTH + 1];	//Add 1 more 0x00 byte for pp_bmc_log_info usage
	static unsigned int asus_mb_sup_mdl_add_cnt = 0;
	unsigned char add_idx = 0;

    memset(&models, 0, sizeof(models));
    // scan args
    if (pp_tp_arg_scanf(args, 0, &err, "s|sssssssssssssss",
    			&(models[0]), &(models[1]), &(models[2]), &(models[3]),
    			&(models[4]), &(models[5]), &(models[6]), &(models[7]),
    			&(models[8]), &(models[9]), &(models[10]), &(models[11]),
    			&(models[12]), &(models[13]), &(models[14]), &(models[15])) != 16) {
        pp_bmc_log_perror("%s (%s) failed: %s ", cname, id, pp_strstream_buf(&err));
        goto finally_mb_sup_mdl;
    }

//    base = malloc(sizeof(pp_tp_obj_t));
//    pp_tp_obj_init(base, PP_TP_SDR, id, pp_asus_mb_support_model_dtor);

	if (asus_mb_sup_mdl_add_cnt < ASUS_MB_SUPPORT_MODEL_MAX_NUMBER) {
	    /* open file */
	    assert(ASUS_MB_SUPPORT_MODEL_FILE_PATH);
	    if (0 > (fd = open(ASUS_MB_SUPPORT_MODEL_FILE_PATH, O_WRONLY | O_CREAT | O_APPEND, S_RW_RW_RW_))) {
	        pp_bmc_log_perror("%s can't open file %s:", cname, ASUS_MB_SUPPORT_MODEL_FILE_PATH);
	        goto finally_mb_sup_mdl;
	    };
		do {
			memset(mdl_string, 0, sizeof(mdl_string));
			if (strlen(models[add_idx]) > ASUS_MB_SUPPORT_MODEL_MAX_LENGTH) {
				memcpy(mdl_string, models[add_idx], ASUS_MB_SUPPORT_MODEL_MAX_LENGTH);
			} else {
				memcpy(mdl_string, models[add_idx], strlen(models[add_idx]));
			};
		    /* write file */
		    n = write(fd, mdl_string, ASUS_MB_SUPPORT_MODEL_MAX_LENGTH);
		    if ((n < 0) || (n != ASUS_MB_SUPPORT_MODEL_MAX_LENGTH)) {
		        pp_bmc_log_error("%s write file %s error", cname, ASUS_MB_SUPPORT_MODEL_FILE_PATH);
		        close(fd);
		        goto finally_mb_sup_mdl;
		    };
		    ++asus_mb_sup_mdl_add_cnt;
	        pp_bmc_log_info("%s Entry add to file %s #%d '%s'",
	        		cname, ASUS_MB_SUPPORT_MODEL_FILE_PATH, asus_mb_sup_mdl_add_cnt, mdl_string);
	    } while ((asus_mb_sup_mdl_add_cnt < ASUS_MB_SUPPORT_MODEL_MAX_NUMBER) && (models[++add_idx] != NULL));
        close(fd);
	};

 finally_mb_sup_mdl:
    pp_strstream_free(&err);
    return base;
}

//static void pp_asus_mb_support_model_dtor(pp_tp_obj_t* o) {    
//    free (o);
//}


//	The ASUS POST INIT for message LED OFF IPMI OEM cmd value in file adding.
pp_tp_obj_t* pp_asus_bmc_post_init_msgled_value_ctor(const char* id, vector_t* args) {
    pp_tp_obj_t *base = NULL;
    const char *cname = "[ASUSPostInitMsgLedValue]";
    pp_strstream_t err = PP_STRSTREAM_INITIALIZER;
    int fd, n;
    unsigned char value;

    // scan args
    if (pp_tp_arg_scanf(args, 0, &err, "d", &value) != 1) {
        pp_bmc_log_perror("%s (%s) failed: %s ", cname, id, pp_strstream_buf(&err));
        goto finally_asus_bmc_post_init_msgled_value;
    }

//    base = malloc(sizeof(pp_tp_obj_t));
//    pp_tp_obj_init(base, PP_TP_SDR, id, pp_asus_bmc_post_init_msgled_value_dtor);

    /* open file */
    assert(ASUS_BMC_POST_INIT_MSGLED_VALUE_FILE_PATH);
    if (0 > (fd = open(ASUS_BMC_POST_INIT_MSGLED_VALUE_FILE_PATH, O_WRONLY | O_CREAT | O_APPEND, S_RW_RW_RW_))) {
        pp_bmc_log_perror("%s can't open file %s:", cname, ASUS_BMC_POST_INIT_MSGLED_VALUE_FILE_PATH);
        goto finally_asus_bmc_post_init_msgled_value;
    };
    /* write file */
    n = write(fd, &value, 1);
    if ((n < 0) || (n != 1)) {
        pp_bmc_log_error("%s write file %s error", cname, ASUS_BMC_POST_INIT_MSGLED_VALUE_FILE_PATH);
        close(fd);
        goto finally_asus_bmc_post_init_msgled_value;
    };
	pp_bmc_log_info("%s Value write to file %s  '0x%2.2X'", cname, ASUS_BMC_POST_INIT_MSGLED_VALUE_FILE_PATH, value);
	close(fd);

 finally_asus_bmc_post_init_msgled_value:
    pp_strstream_free(&err);
    return base;
}

//static void pp_asus_bmc_post_init_msgled_value_dtor(pp_tp_obj_t* o) {    
//    free (o);
//}



