/**
 * blackford_mch_sensor.c
 *
 * implements all BLACKFORD_MCH sensors
 * 
 */

#include <malloc.h>

#include <pp/base.h>
#include <pp/bmc/ipmi_sdr.h>
#include <pp/bmc/ipmi_bits.h>

#include <pp/bmc/ipmi_sdr.h>
#include <pp/bmc/topo_base_obj.h>
#include <pp/bmc/topo_factory.h>
#include <pp/bmc/tp_i2c_comdev.h>
#include <pp/bmc/tp_scan_sensdev.h>
#include <pp/bmc/bmc_config.h>
#include <pp/bmc/debug.h>

#include <pp/i2c.h>

#include "drivers/blackford_mch_sensor.h"
#include "drivers/blackford_mch_chip.h"

#define RDO_chk_mask		0x0080
#define WOD_chk_mask		0x0040
#define SBE_chk_mask		0x0020
#define BUSY_chk_mask		0x0010
#define DATA_chk_mask		0xFF00

#define SMBus_cfg_read_dword_word_begin			0x81
#define SMBus_cfg_read_dword_word_medium		0x01
#define SMBus_cfg_read_dword_word_end			0x41
#define SMBus_cfg_read_dword_byte_begin			0x80
#define SMBus_cfg_read_dword_byte_medium		0x00
#define SMBus_cfg_read_dword_byte_end			0x40
#define SMBus_cfg_write_dword_word_begin		0x8d
#define SMBus_cfg_write_dword_word_medium		0x0d
#define SMBus_cfg_write_dword_word_end			0x4d
#define SMBus_cfg_write_dword_byte_begin		0x8c
#define SMBus_cfg_write_dword_byte_medium		0x0c
#define SMBus_cfg_write_dword_byte_end			0x4c

#define I2C_1_NUMBER	1
#define I2C_2_NUMBER	2
#define I2C_3_NUMBER	3
#define I2C_4_NUMBER	4
#define I2C_5_NUMBER	5

/*
 * BLACKFORD_MCH sensor structure
 */
typedef struct sensor_config_s {
    /* the sensor type */
    unsigned char sensor_type;
    unsigned char reading_type;
    unsigned char sdr_type;
    unsigned char unit_type;
    unsigned char analog_format;
    /* update function */
    void (*update)(pp_sensor_scannable_t*);
    /* configuration */
    struct {
	    BITFIELD2(unsigned char, channel : 4, slot : 4);
		unsigned char pci_bus;
		unsigned char pci_device;
		unsigned char pci_function;
		unsigned char pci_register_status;	//2-byte status
		unsigned char pci_register_spdcmd;	//4-byte spd command
    } config; 
    /* sdr configuration */
    unsigned short sdr_mtol;
    unsigned short sdr_bacc;
    char sdr_bexp;
    char sdr_rexp;
} sensor_config_t;

typedef struct {
    /* the basic sensor description, must be first */
    pp_tp_scan_sensdev_t base;

    /* which blackford chip to use */
    pp_tp_i2c_chip_t* i2cchip;

    /* sensor configuration */
    sensor_config_t* cfg;
} blackford_mch_sensor_t;

/* internal prototypes */
static blackford_mch_sensor_t* blackford_mch_sensor_create(/* globals */
					       const char* id,
					       pp_tp_i2c_chip_t* i2cchip,
					       int sensor_idx,
					       pp_tp_cond_t* scan);
//static char* blackford_mch_to_string(blackford_mch_sensor_t* sensor);
static ipmi_sdr_header_t* blackford_mch_default_sdr(pp_tp_sensdev_t* s);
static void blackford_mch_update_temp_reading(pp_sensor_scannable_t* s);
static int blackford_mch_update_temp_status_check(unsigned short status_value, unsigned short mask);
static void blackford_mch_destroy(blackford_mch_sensor_t* sensor);


/*
 * private structures
 ************************/
static sensor_config_t sensor_config[] = {
    { /* 0: channel 0, slot 0 */
        .sensor_type = IPMI_SENSOR_TYPE_TEMPERATURE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_DEGREES_C,
	.analog_format = IPMI_ANALOG_DATA_FORMAT_2_COMPL,
	.update = blackford_mch_update_temp_reading,
        .config = { .channel = 0, .slot = 0, .pci_bus = 0, .pci_device=21 , .pci_function = 0,
        				.pci_register_status = 0x74, .pci_register_spdcmd = 0x78},
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(1, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0), 
	.sdr_bexp = 0, .sdr_rexp = 0,
    }, { /* 1: channel 0, slot 1 */
        .sensor_type = IPMI_SENSOR_TYPE_TEMPERATURE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_DEGREES_C,
	.analog_format = IPMI_ANALOG_DATA_FORMAT_2_COMPL,
	.update = blackford_mch_update_temp_reading,
        .config = { .channel = 0, .slot = 1, .pci_bus = 0, .pci_device=21 , .pci_function = 0,
        				.pci_register_status = 0x74, .pci_register_spdcmd = 0x78},
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(1, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0), 
	.sdr_bexp = 0, .sdr_rexp = 0,
    }, { /* 2: channel 0, slot 2 */
        .sensor_type = IPMI_SENSOR_TYPE_TEMPERATURE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_DEGREES_C,
	.analog_format = IPMI_ANALOG_DATA_FORMAT_2_COMPL,
	.update = blackford_mch_update_temp_reading,
        .config = { .channel = 0, .slot = 2, .pci_bus = 0, .pci_device=21 , .pci_function = 0,
        				.pci_register_status = 0x74, .pci_register_spdcmd = 0x78},
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(1, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0), 
	.sdr_bexp = 0, .sdr_rexp = 0,
    }, { /* 3: channel 0, slot 3 */
        .sensor_type = IPMI_SENSOR_TYPE_TEMPERATURE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_DEGREES_C,
	.analog_format = IPMI_ANALOG_DATA_FORMAT_2_COMPL,
	.update = blackford_mch_update_temp_reading,
        .config = { .channel = 0, .slot = 3, .pci_bus = 0, .pci_device=21 , .pci_function = 0,
        				.pci_register_status = 0x74, .pci_register_spdcmd = 0x78},
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(1, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0), 
	.sdr_bexp = 0, .sdr_rexp = 0,
    }, { /* 4: channel 1, slot 0 */
        .sensor_type = IPMI_SENSOR_TYPE_TEMPERATURE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_DEGREES_C,
	.analog_format = IPMI_ANALOG_DATA_FORMAT_2_COMPL,
	.update = blackford_mch_update_temp_reading,
        .config = { .channel = 1, .slot = 0, .pci_bus = 0, .pci_device=21 , .pci_function = 0,
        				.pci_register_status = 0x76, .pci_register_spdcmd = 0x7c},
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(1, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0), 
	.sdr_bexp = 0, .sdr_rexp = 0,
    }, { /* 5: channel 1, slot 1 */
        .sensor_type = IPMI_SENSOR_TYPE_TEMPERATURE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_DEGREES_C,
	.analog_format = IPMI_ANALOG_DATA_FORMAT_2_COMPL,
	.update = blackford_mch_update_temp_reading,
        .config = { .channel = 1, .slot = 1, .pci_bus = 0, .pci_device=21 , .pci_function = 0,
        				.pci_register_status = 0x76, .pci_register_spdcmd = 0x7c},
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(1, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0), 
	.sdr_bexp = 0, .sdr_rexp = 0,
    }, { /* 6: channel 1, slot 2 */
        .sensor_type = IPMI_SENSOR_TYPE_TEMPERATURE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_DEGREES_C,
	.analog_format = IPMI_ANALOG_DATA_FORMAT_2_COMPL,
	.update = blackford_mch_update_temp_reading,
        .config = { .channel = 1, .slot = 2, .pci_bus = 0, .pci_device=21 , .pci_function = 0,
        				.pci_register_status = 0x76, .pci_register_spdcmd = 0x7c},
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(1, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0), 
	.sdr_bexp = 0, .sdr_rexp = 0,
    }, { /* 7: channel 1, slot 3 */
        .sensor_type = IPMI_SENSOR_TYPE_TEMPERATURE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_DEGREES_C,
	.analog_format = IPMI_ANALOG_DATA_FORMAT_2_COMPL,
	.update = blackford_mch_update_temp_reading,
        .config = { .channel = 1, .slot = 3, .pci_bus = 0, .pci_device=21 , .pci_function = 0,
        				.pci_register_status = 0x76, .pci_register_spdcmd = 0x7c},
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(1, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0), 
	.sdr_bexp = 0, .sdr_rexp = 0,
    }, { /* 8: channel 2, slot 0 */
        .sensor_type = IPMI_SENSOR_TYPE_TEMPERATURE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_DEGREES_C,
	.analog_format = IPMI_ANALOG_DATA_FORMAT_2_COMPL,
	.update = blackford_mch_update_temp_reading,
        .config = { .channel = 2, .slot = 0, .pci_bus = 0, .pci_device=22 , .pci_function = 0,
        				.pci_register_status = 0x74, .pci_register_spdcmd = 0x78},
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(1, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0), 
	.sdr_bexp = 0, .sdr_rexp = 0,
    }, { /* 9: channel 2, slot 1 */
        .sensor_type = IPMI_SENSOR_TYPE_TEMPERATURE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_DEGREES_C,
	.analog_format = IPMI_ANALOG_DATA_FORMAT_2_COMPL,
	.update = blackford_mch_update_temp_reading,
        .config = { .channel = 2, .slot = 1, .pci_bus = 0, .pci_device=22 , .pci_function = 0,
        				.pci_register_status = 0x74, .pci_register_spdcmd = 0x78},
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(1, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0), 
	.sdr_bexp = 0, .sdr_rexp = 0,
    }, { /* 10: channel 2, slot 2 */
        .sensor_type = IPMI_SENSOR_TYPE_TEMPERATURE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_DEGREES_C,
	.analog_format = IPMI_ANALOG_DATA_FORMAT_2_COMPL,
	.update = blackford_mch_update_temp_reading,
        .config = { .channel = 2, .slot = 2, .pci_bus = 0, .pci_device=22 , .pci_function = 0,
        				.pci_register_status = 0x74, .pci_register_spdcmd = 0x78},
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(1, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0), 
	.sdr_bexp = 0, .sdr_rexp = 0,
    }, { /* 11: channel 2, slot 3 */
        .sensor_type = IPMI_SENSOR_TYPE_TEMPERATURE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_DEGREES_C,
	.analog_format = IPMI_ANALOG_DATA_FORMAT_2_COMPL,
	.update = blackford_mch_update_temp_reading,
        .config = { .channel = 2, .slot = 3, .pci_bus = 0, .pci_device=22 , .pci_function = 0,
        				.pci_register_status = 0x74, .pci_register_spdcmd = 0x78},
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(1, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0), 
	.sdr_bexp = 0, .sdr_rexp = 0,
    }, { /* 12: channel 3, slot 0 */
        .sensor_type = IPMI_SENSOR_TYPE_TEMPERATURE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_DEGREES_C,
	.analog_format = IPMI_ANALOG_DATA_FORMAT_2_COMPL,
	.update = blackford_mch_update_temp_reading,
        .config = { .channel = 3, .slot = 0, .pci_bus = 0, .pci_device=22 , .pci_function = 0,
        				.pci_register_status = 0x76, .pci_register_spdcmd = 0x7c},
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(1, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0), 
	.sdr_bexp = 0, .sdr_rexp = 0,
    }, { /* 13: channel 3, slot 1 */
        .sensor_type = IPMI_SENSOR_TYPE_TEMPERATURE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_DEGREES_C,
	.analog_format = IPMI_ANALOG_DATA_FORMAT_2_COMPL,
	.update = blackford_mch_update_temp_reading,
        .config = { .channel = 3, .slot = 1, .pci_bus = 0, .pci_device=22 , .pci_function = 0,
        				.pci_register_status = 0x76, .pci_register_spdcmd = 0x7c},
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(1, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0), 
	.sdr_bexp = 0, .sdr_rexp = 0,
    }, { /* 14: channel 3, slot 2 */
        .sensor_type = IPMI_SENSOR_TYPE_TEMPERATURE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_DEGREES_C,
	.analog_format = IPMI_ANALOG_DATA_FORMAT_2_COMPL,
	.update = blackford_mch_update_temp_reading,
        .config = { .channel = 3, .slot = 2, .pci_bus = 0, .pci_device=22 , .pci_function = 0,
        				.pci_register_status = 0x76, .pci_register_spdcmd = 0x7c},
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(1, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0), 
	.sdr_bexp = 0, .sdr_rexp = 0,
    }, { /* 15: channel 3, slot 3 */
        .sensor_type = IPMI_SENSOR_TYPE_TEMPERATURE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_DEGREES_C,
	.analog_format = IPMI_ANALOG_DATA_FORMAT_2_COMPL,
	.update = blackford_mch_update_temp_reading,
        .config = { .channel = 3, .slot = 3, .pci_bus = 0, .pci_device=22 , .pci_function = 0,
        				.pci_register_status = 0x76, .pci_register_spdcmd = 0x7c},
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(1, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0), 
	.sdr_bexp = 0, .sdr_rexp = 0,
	},
};



/*
 * implementations
 */

pp_tp_obj_t* pp_blackford_mch_sensor_ctor(const char* id, vector_t* args) {
    pp_strstream_t err = PP_STRSTREAM_INITIALIZER;
    blackford_mch_sensor_t* o;

    pp_tp_i2c_chip_t* i2cchip;
    int snum;
    pp_tp_cond_t* scan;
    
    if (pp_tp_arg_scanf(args, 0, &err, "o<h>d|o<sc>", &i2cchip, &snum, &scan) != 3)
    {
		pp_bmc_log_error("[BLACKFORD_MCH] sensor '%s' initialization failed: %s", id, pp_strstream_buf(&err));
		o = NULL;
    } else if (snum > PP_SENSOR_BLACKFORD_MCH_MAX_SENSOR_NUM) {
		pp_bmc_log_error("[BLACKFORD_MCH] sensor '%s' initialization: invalid sensor number: %d", id, snum);
		o = NULL;
    } else if (strcmp((i2cchip)->model, "blackford_mch") != 0) {
        pp_bmc_log_error("[BLACKFORD_MCH] sensor '%s' needs blackford_mch chip to initialize", id);
        o = NULL;
    } else{
        o = blackford_mch_sensor_create(id, i2cchip, snum, scan);
    }
    return (pp_tp_obj_t*)o;
}						

/*
 * private functions
 ********************************************************************/

static blackford_mch_sensor_t* blackford_mch_sensor_create(/* globals */
					      const char* id,
					      pp_tp_i2c_chip_t* i2cchip,
					      int sensor_idx,
					      pp_tp_cond_t* scan)
{
    sensor_config_t *s = &sensor_config[sensor_idx];
    blackford_mch_sensor_t* blackfords = malloc(sizeof(blackford_mch_sensor_t));

    /* globals */
    pp_tp_scan_sensdev_init(&blackfords->base, PP_TP_SCAN_SENS_DEV, id,
			    (void(*)(pp_tp_obj_t*))blackford_mch_destroy,
			    blackford_mch_default_sdr, s->update, scan);
    blackfords->i2cchip = pp_tp_i2c_chip_duplicate(i2cchip);
    blackfords->cfg = s;
	
    /* register for scanning  */
    pp_tp_scan_sensdev_register(&blackfords->base, 0);

    return blackfords;
}

static ipmi_sdr_header_t* blackford_mch_default_sdr(pp_tp_sensdev_t* s) {
    sensor_config_t* cfg = ((blackford_mch_sensor_t*)s)->cfg;
    ipmi_sdr_header_t* mysdr = NULL;

    if (cfg->sdr_type == IPMI_SDR_FULL_SENSOR) {
	    mysdr = (ipmi_sdr_header_t*)ipmi_sdr_full_part_create(cfg->sensor_type,
					cfg->reading_type,
			       cfg->unit_type,
			       cfg->analog_format,
			       cfg->sdr_mtol, cfg->sdr_bacc,
			       cfg->sdr_bexp, cfg->sdr_rexp,
			       ipmi_sdr_max_by_form(cfg->analog_format),
			       ipmi_sdr_min_by_form(cfg->analog_format));
    } else if (cfg->sdr_type == IPMI_SDR_COMPACT_SENSOR) {
		mysdr = (ipmi_sdr_header_t*)ipmi_sdr_full_part_create(
	    cfg->sensor_type, cfg->reading_type, 0, 0, 0, 0, 0, 0, 0, 0);
    } else {
		pp_bmc_log_error("blackford_mch_default_sdr failed: SDR type %d not suptemped", cfg->sdr_type);
    }
				      
    return (ipmi_sdr_header_t*)mysdr;
}

static void blackford_mch_update_temp_reading(pp_sensor_scannable_t* s) {
    blackford_mch_sensor_t* blackfords = PP_TP_SCANNABLE_2_OBJ_CAST(s, blackford_mch_sensor_t);
    sensor_config_t* cfg = blackfords->cfg;
    pp_tp_i2c_chip_t* i2cchip = blackfords->i2cchip;
    int ret = PP_ERR;
    int smbret = 0;
    int status_check_ret;
    unsigned short buf[2];
    unsigned char smbstatus;
    unsigned char value;
	unsigned short busdevicefun = ( (((cfg->config.pci_device & 0x1F) << 3) | (cfg->config.pci_function & 0x07)) << 8 ) | (cfg->config.pci_bus);
	unsigned short spdstatusreg = ( ((cfg->config.pci_register_status) & 0xFC) << 8 ) | (0x00);
	unsigned short spdcmdreg = ( ((cfg->config.pci_register_spdcmd) & 0xFC) << 8 ) | (0x00);
	
//$$$$$ pp_tp_i2c_chip_tx_word_data : (write lower byte first, then the higher byte)
//$$$$$ pp_tp_i2c_chip_rx_word_data : (read and place in higher byte first, then the lower byte)

    if (pp_tp_i2c_chip_is_initialized(i2cchip) && pp_tp_scan_sensdev_do_scan(&blackfords->base)) {
		if (PP_SUC == (ret = pp_tp_i2c_chip_pre_com(i2cchip))) {
	//SMBus Configuration Read (Word Write/Byte Read/Word Read)
			smbret = pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_read_dword_word_begin, busdevicefun);
		    if (smbret != PP_SUC) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #0-1");
		    	goto ending;
		    };
			smbret = pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_read_dword_word_end, spdstatusreg);
		    if (smbret != PP_SUC) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #0-2");
		    	goto ending;
		    };
			smbret = pp_tp_i2c_chip_rx_byte_data(i2cchip, SMBus_cfg_read_dword_byte_begin, &smbstatus);
		    if (smbret != PP_SUC) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #0-3");
		    	goto ending;
		    };
			smbret = pp_tp_i2c_chip_rx_word_data(i2cchip, SMBus_cfg_read_dword_word_medium, &buf[0]);
		    if (smbret != PP_SUC) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #0-4");
		    	goto ending;
		    };
			smbret = pp_tp_i2c_chip_rx_word_data(i2cchip, SMBus_cfg_read_dword_word_end, &buf[1]);
		    if (smbret != PP_SUC) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #0-5");
		    	goto ending;
		    };
	//status = 1 & BUSY check
		    if (smbstatus != 0x01) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #0-6");
		    	goto ending;
		    };
		    if((cfg->config.channel == 0) || (cfg->config.channel == 2))
				status_check_ret = blackford_mch_update_temp_status_check(buf[1], BUSY_chk_mask);
			else
				status_check_ret = blackford_mch_update_temp_status_check(buf[0], BUSY_chk_mask);
		    if (status_check_ret) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #0-7");
		    	goto ending;
		    };

// ### Step-1 ###  SMBus Configuration Write (Word Write)
			buf[0] = (0x80 << 8) | (0xB8 | (cfg->config.slot & 0x07));
			buf[1] = (0x01 << 8) | (0x00);
			smbret = pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_begin, busdevicefun);
		    if (smbret != PP_SUC) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #1-1");
		    	goto ending;
		    };
			smbret = pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_medium, spdcmdreg);
		    if (smbret != PP_SUC) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #1-2");
		    	goto ending;
		    };
			smbret = pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_medium, buf[0]);
		    if (smbret != PP_SUC) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #1-3");
		    	goto ending;
		    };
			smbret = pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_end, buf[1]);
		    if (smbret != PP_SUC) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #1-4");
		    	goto ending;
		    };
	//SMBus Configuration Read (Word Write/Byte Read/Word Read)
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_read_dword_word_begin, busdevicefun);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_read_dword_word_end, spdstatusreg);
			pp_tp_i2c_chip_rx_byte_data(i2cchip, SMBus_cfg_read_dword_byte_begin, &smbstatus);
			pp_tp_i2c_chip_rx_word_data(i2cchip, SMBus_cfg_read_dword_word_medium, &buf[0]);
			pp_tp_i2c_chip_rx_word_data(i2cchip, SMBus_cfg_read_dword_word_end, &buf[1]);
	//status = 1 & WOD,SBE check
		    if (smbstatus != 0x01) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #1-5");
		    	goto ending;
		    };
		    if((cfg->config.channel == 0) || (cfg->config.channel == 2))
				status_check_ret = blackford_mch_update_temp_status_check(buf[1], SBE_chk_mask | WOD_chk_mask);
			else
				status_check_ret = blackford_mch_update_temp_status_check(buf[0], SBE_chk_mask | WOD_chk_mask);
		    if (status_check_ret) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #1-6");
		    	goto ending;
		    };

// ### Step-2 ###  SMBus Configuration Write (Word Write)
			buf[0] = (0x00 << 8) | (0xB8 | (cfg->config.slot & 0x07));
			buf[1] = (0x01 << 8) | (0x03);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_begin, busdevicefun);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_medium, spdcmdreg);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_medium, buf[0]);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_end, buf[1]);
	//SMBus Configuration Read (Word Write/Byte Read/Word Read)
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_read_dword_word_begin, busdevicefun);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_read_dword_word_end, spdstatusreg);
			pp_tp_i2c_chip_rx_byte_data(i2cchip, SMBus_cfg_read_dword_byte_begin, &smbstatus);
			pp_tp_i2c_chip_rx_word_data(i2cchip, SMBus_cfg_read_dword_word_medium, &buf[0]);
			pp_tp_i2c_chip_rx_word_data(i2cchip, SMBus_cfg_read_dword_word_end, &buf[1]);
	//status = 1 & WOD,SBE check
		    if (smbstatus != 0x01) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #2-1");
		    	goto ending;
		    };
		    if((cfg->config.channel == 0) || (cfg->config.channel == 2))
				status_check_ret = blackford_mch_update_temp_status_check(buf[1], SBE_chk_mask | WOD_chk_mask);
			else
				status_check_ret = blackford_mch_update_temp_status_check(buf[0], SBE_chk_mask | WOD_chk_mask);
		    if (status_check_ret) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #2-2");
		    	goto ending;
		    };

// ### Step-3 ###  SMBus Configuration Write (Word Write)
			buf[0] = (0x00 << 8) | (0xB8 | (cfg->config.slot & 0x07));
			buf[1] = (0x01 << 8) | (0x00);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_begin, busdevicefun);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_medium, spdcmdreg);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_medium, buf[0]);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_end, buf[1]);
	//SMBus Configuration Read (Word Write/Byte Read/Word Read)
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_read_dword_word_begin, busdevicefun);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_read_dword_word_end, spdstatusreg);
			pp_tp_i2c_chip_rx_byte_data(i2cchip, SMBus_cfg_read_dword_byte_begin, &smbstatus);
			pp_tp_i2c_chip_rx_word_data(i2cchip, SMBus_cfg_read_dword_word_medium, &buf[0]);
			pp_tp_i2c_chip_rx_word_data(i2cchip, SMBus_cfg_read_dword_word_end, &buf[1]);
	//status = 1 & WOD,SBE check
		    if (smbstatus != 0x01) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #3-1");
		    	goto ending;
		    };
		    if((cfg->config.channel == 0) || (cfg->config.channel == 2))
				status_check_ret = blackford_mch_update_temp_status_check(buf[1], SBE_chk_mask | WOD_chk_mask);
			else
				status_check_ret = blackford_mch_update_temp_status_check(buf[0], SBE_chk_mask | WOD_chk_mask);
		    if (status_check_ret) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #3-2");
		    	goto ending;
		    };

// ### Step-4 ###  SMBus Configuration Write (Word Write)
			buf[0] = (0x40 << 8) | (0xB8 | (cfg->config.slot & 0x07));
			buf[1] = (0x01 << 8) | (0x84);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_begin, busdevicefun);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_medium, spdcmdreg);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_medium, buf[0]);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_end, buf[1]);
	//SMBus Configuration Read (Word Write/Byte Read/Word Read)
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_read_dword_word_begin, busdevicefun);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_read_dword_word_end, spdstatusreg);
			pp_tp_i2c_chip_rx_byte_data(i2cchip, SMBus_cfg_read_dword_byte_begin, &smbstatus);
			pp_tp_i2c_chip_rx_word_data(i2cchip, SMBus_cfg_read_dword_word_medium, &buf[0]);
			pp_tp_i2c_chip_rx_word_data(i2cchip, SMBus_cfg_read_dword_word_end, &buf[1]);
	//status = 1 & WOD,SBE check
		    if (smbstatus != 0x01) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #4-1");
		    	goto ending;
		    };
		    if((cfg->config.channel == 0) || (cfg->config.channel == 2))
				status_check_ret = blackford_mch_update_temp_status_check(buf[1], SBE_chk_mask | WOD_chk_mask);
			else
				status_check_ret = blackford_mch_update_temp_status_check(buf[0], SBE_chk_mask | WOD_chk_mask);
		    if (status_check_ret) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #4-2");
		    	goto ending;
		    };

// ### Step-5 ###  SMBus Configuration Write (Word Write)
			buf[0] = (0x80 << 8) | (0xB8 | (cfg->config.slot & 0x07));
			buf[1] = (0x00 << 8) | (0x00);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_begin, busdevicefun);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_medium, spdcmdreg);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_medium, buf[0]);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_end, buf[1]);
	//SMBus Configuration Read (Word Write/Byte Read/Word Read)
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_read_dword_word_begin, busdevicefun);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_read_dword_word_end, spdstatusreg);
			pp_tp_i2c_chip_rx_byte_data(i2cchip, SMBus_cfg_read_dword_byte_begin, &smbstatus);
			pp_tp_i2c_chip_rx_word_data(i2cchip, SMBus_cfg_read_dword_word_medium, &buf[0]);
			pp_tp_i2c_chip_rx_word_data(i2cchip, SMBus_cfg_read_dword_word_end, &buf[1]);
	//status = 1 & DATA,RDO,SBE check
		    if (smbstatus != 0x01) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #5-1");
		    	goto ending;
		    };
		    if((cfg->config.channel == 0) || (cfg->config.channel == 2))
				status_check_ret = blackford_mch_update_temp_status_check(buf[1], SBE_chk_mask | RDO_chk_mask | DATA_chk_mask);
			else
				status_check_ret = blackford_mch_update_temp_status_check(buf[0], SBE_chk_mask | RDO_chk_mask | DATA_chk_mask);
		    if (status_check_ret) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #5-2");
		    	goto ending;
		    };

// ### Step-6 ###  SMBus Configuration Write (Word Write)
			buf[0] = (0x00 << 8) | (0xB8 | (cfg->config.slot & 0x07));
			buf[1] = (0x00 << 8) | (0x00);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_begin, busdevicefun);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_medium, spdcmdreg);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_medium, buf[0]);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_end, buf[1]);
	//SMBus Configuration Read (Word Write/Byte Read/Word Read)
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_read_dword_word_begin, busdevicefun);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_read_dword_word_end, spdstatusreg);
			pp_tp_i2c_chip_rx_byte_data(i2cchip, SMBus_cfg_read_dword_byte_begin, &smbstatus);
			pp_tp_i2c_chip_rx_word_data(i2cchip, SMBus_cfg_read_dword_word_medium, &buf[0]);
			pp_tp_i2c_chip_rx_word_data(i2cchip, SMBus_cfg_read_dword_word_end, &buf[1]);
	//status = 1 & RDO,SBE check
		    if (smbstatus != 0x01) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #6-1");
		    	goto ending;
		    };
		    if((cfg->config.channel == 0) || (cfg->config.channel == 2))
				status_check_ret = blackford_mch_update_temp_status_check(buf[1], SBE_chk_mask | RDO_chk_mask);
			else
				status_check_ret = blackford_mch_update_temp_status_check(buf[0], SBE_chk_mask | RDO_chk_mask);
		    if (status_check_ret) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #6-2");
		    	goto ending;
		    };

// ### Step-7 ###  SMBus Configuration Write (Word Write)
			buf[0] = (0x00 << 8) | (0xB8 | (cfg->config.slot & 0x07));
			buf[1] = (0x00 << 8) | (0x00);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_begin, busdevicefun);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_medium, spdcmdreg);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_medium, buf[0]);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_end, buf[1]);
	//SMBus Configuration Read (Word Write/Byte Read/Word Read)
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_read_dword_word_begin, busdevicefun);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_read_dword_word_end, spdstatusreg);
			pp_tp_i2c_chip_rx_byte_data(i2cchip, SMBus_cfg_read_dword_byte_begin, &smbstatus);
			pp_tp_i2c_chip_rx_word_data(i2cchip, SMBus_cfg_read_dword_word_medium, &buf[0]);
			pp_tp_i2c_chip_rx_word_data(i2cchip, SMBus_cfg_read_dword_word_end, &buf[1]);
	//status = 1 & RDO,SBE check
		    if (smbstatus != 0x01) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #7-1");
		    	goto ending;
		    };
		    if((cfg->config.channel == 0) || (cfg->config.channel == 2))
				status_check_ret = blackford_mch_update_temp_status_check(buf[1], SBE_chk_mask | RDO_chk_mask);
			else
				status_check_ret = blackford_mch_update_temp_status_check(buf[0], SBE_chk_mask | RDO_chk_mask);
		    if (status_check_ret) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #7-2");
		    	goto ending;
		    };

// ### Step-8 ###  SMBus Configuration Write (Word Write)
			buf[0] = (0x00 << 8) | (0xB8 | (cfg->config.slot & 0x07));
			buf[1] = (0x00 << 8) | (0x00);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_begin, busdevicefun);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_medium, spdcmdreg);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_medium, buf[0]);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_end, buf[1]);
	//SMBus Configuration Read (Word Write/Byte Read/Word Read)
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_read_dword_word_begin, busdevicefun);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_read_dword_word_end, spdstatusreg);
			pp_tp_i2c_chip_rx_byte_data(i2cchip, SMBus_cfg_read_dword_byte_begin, &smbstatus);
			pp_tp_i2c_chip_rx_word_data(i2cchip, SMBus_cfg_read_dword_word_medium, &buf[0]);
			pp_tp_i2c_chip_rx_word_data(i2cchip, SMBus_cfg_read_dword_word_end, &buf[1]);
	//status = 1 & RDO,SBE check
		    if (smbstatus != 0x01) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #8-1");
		    	goto ending;
		    };
		    if((cfg->config.channel == 0) || (cfg->config.channel == 2))
				status_check_ret = blackford_mch_update_temp_status_check(buf[1], SBE_chk_mask | RDO_chk_mask);
			else
				status_check_ret = blackford_mch_update_temp_status_check(buf[0], SBE_chk_mask | RDO_chk_mask);
		    if (status_check_ret) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #8-2");
		    	goto ending;
		    };
//*****  Raw AMB temperature value
		    if((cfg->config.channel == 0) || (cfg->config.channel == 2))
				value = (buf[1] & 0xFF00) >> 9;
			else
				value = (buf[0] & 0xFF00) >> 9;

// ### Step-8 ###  SMBus Configuration Write (Word Write)
			buf[0] = (0x40 << 8) | (0xB8 | (cfg->config.slot & 0x07));
			buf[1] = (0x00 << 8) | (0x00);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_begin, busdevicefun);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_medium, spdcmdreg);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_medium, buf[0]);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_write_dword_word_end, buf[1]);
	//SMBus Configuration Read (Word Write/Byte Read/Word Read)
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_read_dword_word_begin, busdevicefun);
			pp_tp_i2c_chip_tx_word_data(i2cchip, SMBus_cfg_read_dword_word_end, spdstatusreg);
			pp_tp_i2c_chip_rx_byte_data(i2cchip, SMBus_cfg_read_dword_byte_begin, &smbstatus);
			pp_tp_i2c_chip_rx_word_data(i2cchip, SMBus_cfg_read_dword_word_medium, &buf[0]);
			pp_tp_i2c_chip_rx_word_data(i2cchip, SMBus_cfg_read_dword_word_end, &buf[1]);
	//status = 1 & RDO,SBE check
		    if (smbstatus != 0x01) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #9-1");
		    	goto ending;
		    };
		    if((cfg->config.channel == 0) || (cfg->config.channel == 2))
				status_check_ret = blackford_mch_update_temp_status_check(buf[1], SBE_chk_mask | RDO_chk_mask);
			else
				status_check_ret = blackford_mch_update_temp_status_check(buf[0], SBE_chk_mask | RDO_chk_mask);
		    if (status_check_ret) {
			    pp_bmc_log_debug("[Blackford-sensor] Update AMB temperatur routine fail #9-2");
		    	goto ending;
		    };

			pp_bmc_log_debug("[Blackford-sensor] Get AMB temperature = %d", value);
			ret = value;
ending:		pp_tp_i2c_chip_post_com(i2cchip);
	    } else {
	    	errno = EIO;
	    };
    } else {
	};

    if ( ret == 0 )
       ret = -2;
    pp_tp_scan_sensdev_update_reading(&blackfords->base, ret);	
}

static int blackford_mch_update_temp_status_check(unsigned short status_value, unsigned short mask)
{
	int ret = -1;
	
	if((mask & RDO_chk_mask) == RDO_chk_mask) {
		if((status_value & RDO_chk_mask) != RDO_chk_mask) {
			return ret;
		};
	};
	if((mask & WOD_chk_mask) == WOD_chk_mask) {
		if((status_value & WOD_chk_mask) != WOD_chk_mask) {
			return ret;
		};
	};
	if((mask & SBE_chk_mask) == SBE_chk_mask) {
		if((status_value & SBE_chk_mask) == SBE_chk_mask) {
			return ret;
		};
	};
	if((mask & BUSY_chk_mask) == BUSY_chk_mask) {
		if((status_value & BUSY_chk_mask) == BUSY_chk_mask) {
			return ret;
		};
	};
	if((mask & DATA_chk_mask) == DATA_chk_mask) {
		if((status_value & DATA_chk_mask) != 0x0100) {
			return ret;
		};
	};
	return (0);
}

static void blackford_mch_destroy(blackford_mch_sensor_t* sensor) {
    /* unregister from scanning  */
    pp_tp_scan_sensdev_unregister(&sensor->base);
    
    pp_tp_i2c_chip_release(sensor->i2cchip);
    pp_tp_scan_sensdev_cleanup(&sensor->base);
    free(sensor);
}
