/**
 * pcf8574_sensor.c
 *
 * implements all PCF8574 sensors
 * 
 */

#include <malloc.h>

#include <pp/base.h>
#include <pp/bmc/ipmi_sdr.h>
#include <pp/bmc/ipmi_bits.h>

#include <pp/bmc/ipmi_sdr.h>
#include <pp/bmc/topo_base_obj.h>
#include <pp/bmc/topo_factory.h>
#include <pp/bmc/tp_i2c_comdev.h>
#include <pp/bmc/tp_scan_sensdev.h>
#include <pp/bmc/bmc_config.h>
#include <pp/bmc/debug.h>

#include "drivers/pcf8574_sensor.h"
#include "drivers/pcf8574_chip.h"



/*
 * PCF8574 sensor structure
 */
typedef struct sensor_config_s {
    /* the sensor type */
    unsigned char sensor_type;
    unsigned char reading_type;
    unsigned char sdr_type;
    unsigned char unit_type;
    unsigned char analog_format;
    /* update function */
    void (*update)(pp_sensor_scannable_t*);
    /* configuration */
    struct {
	    BITFIELD5(unsigned char, bit_shift : 3, bit_in_value : 1, bit_raw_return : 1, all_raw_return : 1, reserved : 2);
	    				//bit_in_value(bit_on_value) = 0 or 1 with meanning of harddisk plugged in or device ON
    } config; 
    /* sdr configuration */
    unsigned short sdr_mtol;
    unsigned short sdr_bacc;
    char sdr_bexp;
    char sdr_rexp;
} sensor_config_t;

typedef struct {
    /* the basic sensor description, must be first */
    pp_tp_scan_sensdev_t base;

    /* which pcf chip to use */
    pp_tp_i2c_chip_t* i2cchip;

    /* sensor configuration */
    sensor_config_t* cfg;
} pcf8574_sensor_t;

/* internal prototypes */
static pcf8574_sensor_t* pcf8574_sensor_create(/* globals */
					       const char* id,
					       pp_tp_i2c_chip_t* i2cchip,
					       int sensor_idx,
					       pp_tp_cond_t* scan);
//static char* pcf8574_to_string(pcf8574_sensor_t* sensor);
static ipmi_sdr_header_t* pcf8574_default_sdr(pp_tp_sensdev_t* s);
static void pcf8574_update_port_reading(pp_sensor_scannable_t* s);
static void pcf8574_destroy(pcf8574_sensor_t* sensor);


/*
 * private structures
 ************************/
static sensor_config_t sensor_config[] = {
    { /* 0 - P0  	(HD plug in/out) */
        .sensor_type = 0xd3,
	.reading_type = IPMI_EVENT_READING_TYPE_SENSOR_SPECIFIC,
	.sdr_type = IPMI_SDR_COMPACT_SENSOR,
	.update = pcf8574_update_port_reading,
        .config = { .bit_shift = 0, .bit_in_value = 1, .bit_raw_return = 0 },
    }, { /* 1 - P1  	(HD plug in/out) */
        .sensor_type = 0xd3,
	.reading_type = IPMI_EVENT_READING_TYPE_SENSOR_SPECIFIC,
	.sdr_type = IPMI_SDR_COMPACT_SENSOR,
	.update = pcf8574_update_port_reading,
        .config = { .bit_shift = 1, .bit_in_value = 1, .bit_raw_return = 0 },
    }, { /* 2 - P2  	(HD plug in/out) */
        .sensor_type = 0xd3,
	.reading_type = IPMI_EVENT_READING_TYPE_SENSOR_SPECIFIC,
	.sdr_type = IPMI_SDR_COMPACT_SENSOR,
	.update = pcf8574_update_port_reading,
        .config = { .bit_shift = 2, .bit_in_value = 1, .bit_raw_return = 0 },
    }, { /* 3 - P3  	(HD plug in/out) */
        .sensor_type = 0xd3,
	.reading_type = IPMI_EVENT_READING_TYPE_SENSOR_SPECIFIC,
	.sdr_type = IPMI_SDR_COMPACT_SENSOR,
	.update = pcf8574_update_port_reading,
        .config = { .bit_shift = 3, .bit_in_value = 1, .bit_raw_return = 0 },
    }, { /* 4 - P4  	(HD plug in/out) */
        .sensor_type = 0xd3,
	.reading_type = IPMI_EVENT_READING_TYPE_SENSOR_SPECIFIC,
	.sdr_type = IPMI_SDR_COMPACT_SENSOR,
	.update = pcf8574_update_port_reading,
        .config = { .bit_shift = 4, .bit_in_value = 1, .bit_raw_return = 0 },
    }, { /* 5 - P5  	(HD plug in/out) */
        .sensor_type = 0xd3,
	.reading_type = IPMI_EVENT_READING_TYPE_SENSOR_SPECIFIC,
	.sdr_type = IPMI_SDR_COMPACT_SENSOR,
	.update = pcf8574_update_port_reading,
        .config = { .bit_shift = 5, .bit_in_value = 1, .bit_raw_return = 0 },
    }, { /* 6 - P6  	(HD plug in/out) */
        .sensor_type = 0xd3,
	.reading_type = IPMI_EVENT_READING_TYPE_SENSOR_SPECIFIC,
	.sdr_type = IPMI_SDR_COMPACT_SENSOR,
	.update = pcf8574_update_port_reading,
        .config = { .bit_shift = 6, .bit_in_value = 1, .bit_raw_return = 0 },
    }, { /* 7 - P7  	(HD plug in/out) */
        .sensor_type = 0xd3,
	.reading_type = IPMI_EVENT_READING_TYPE_SENSOR_SPECIFIC,
	.sdr_type = IPMI_SDR_COMPACT_SENSOR,
	.update = pcf8574_update_port_reading,
        .config = { .bit_shift = 7, .bit_in_value = 1, .bit_raw_return = 0 },
    }, { /* 8 - all 	(LED green/red control) */
        .sensor_type = 0xc1,
	.reading_type = IPMI_EVENT_READING_TYPE_SENSOR_SPECIFIC,
	.sdr_type = IPMI_SDR_COMPACT_SENSOR,
	.update = pcf8574_update_port_reading,
        .config = { .all_raw_return = 1 },
    },
};



/*
 * implementations
 */

pp_tp_obj_t* pp_pcf8574_sensor_ctor(const char* id, vector_t* args) {
    pp_strstream_t err = PP_STRSTREAM_INITIALIZER;
    pcf8574_sensor_t* o=NULL;

    pp_tp_i2c_chip_t* i2cchip;
    int snum;
    pp_tp_cond_t* scan;
    unsigned char sensor_type=0;
    unsigned char reading_type=0;
    unsigned char bit_in_value;
    unsigned char bit_raw_return;
    sensor_config_t *s;
    
    if (pp_tp_arg_scanf(args, 0, &err, "o<h>d|o<sc>", &i2cchip, &snum, &scan) != 3)
    {
		pp_bmc_log_error("[PCF8574] sensor '%s' initialization failed: %s", id, pp_strstream_buf(&err));
		o = NULL;
    } else if (vector_size(args) > 3 && pp_tp_arg_scanf(args, 3, &err, "d<c>d<c>d<c>d<c>",
    	&sensor_type, &reading_type, &bit_in_value, &bit_raw_return) != 4) {
		pp_bmc_log_error("[PCF8574] sensor '%s' initialization: invalid extended arguments", id);
    } else if (snum > PP_SENSOR_PCF8574_MAX_SENSOR_NUM) {
		pp_bmc_log_error("[PCF8574] sensor '%s' initialization: invalid sensor number: %d", id, snum);
		o = NULL;
    } else if (strcmp((i2cchip)->model, "pcf8574") != 0) {
        pp_bmc_log_error("[PCF8574] sensor '%s' needs pcf8574 chip to initialize", id);
        o = NULL;
    } else {
    	if ( sensor_type != 0 && reading_type != 0 ) {
    		s = &sensor_config[snum];
    		s->sensor_type = sensor_type;
    		s->reading_type = reading_type;
    		s->config.bit_in_value = bit_in_value;
    		s->config.bit_raw_return = bit_raw_return;
    	};
        o = pcf8574_sensor_create(id, i2cchip, snum, scan);
    }
    return (pp_tp_obj_t*)o;
}						

/*
 * private functions
 ********************************************************************/

static pcf8574_sensor_t* pcf8574_sensor_create(/* globals */
					      const char* id,
					      pp_tp_i2c_chip_t* i2cchip,
					      int sensor_idx,
					      pp_tp_cond_t* scan)
{
    sensor_config_t *s = &sensor_config[sensor_idx];
    pcf8574_sensor_t* pcfs = malloc(sizeof(pcf8574_sensor_t));

    /* globals */
    pp_tp_scan_sensdev_init(&pcfs->base, PP_TP_SCAN_SENS_DEV, id,
			    (void(*)(pp_tp_obj_t*))pcf8574_destroy,
			    pcf8574_default_sdr, s->update, scan);
    pcfs->i2cchip = pp_tp_i2c_chip_duplicate(i2cchip);
    pcfs->cfg = s;
	
    /* register for scanning  */
    pp_tp_scan_sensdev_register(&pcfs->base, 0);

    return pcfs;
}

static ipmi_sdr_header_t* pcf8574_default_sdr(pp_tp_sensdev_t* s) {
    sensor_config_t* cfg = ((pcf8574_sensor_t*)s)->cfg;
    ipmi_sdr_header_t* mysdr = NULL;

    if (cfg->sdr_type == IPMI_SDR_FULL_SENSOR) {
	    mysdr = (ipmi_sdr_header_t*)ipmi_sdr_full_part_create(cfg->sensor_type,
					cfg->reading_type,
			       cfg->unit_type,
			       cfg->analog_format,
			       cfg->sdr_mtol, cfg->sdr_bacc,
			       cfg->sdr_bexp, cfg->sdr_rexp,
			       ipmi_sdr_max_by_form(cfg->analog_format),
			       ipmi_sdr_min_by_form(cfg->analog_format));
    } else if (cfg->sdr_type == IPMI_SDR_COMPACT_SENSOR) {
		mysdr = (ipmi_sdr_header_t*)ipmi_sdr_full_part_create(
	    cfg->sensor_type, cfg->reading_type, 0, 0, 0, 0, 0, 0, 0, 0);
    } else {
		pp_bmc_log_error("pcf8574_default_sdr failed: SDR type %d not supported", cfg->sdr_type);
    }
				      
    return (ipmi_sdr_header_t*)mysdr;
}

static void pcf8574_update_port_reading(pp_sensor_scannable_t* s) {
    pcf8574_sensor_t* pcfs = PP_TP_SCANNABLE_2_OBJ_CAST(s, pcf8574_sensor_t);
    sensor_config_t* cfg = pcfs->cfg;
    pp_tp_i2c_chip_t* i2cchip = pcfs->i2cchip;
    int ret = PP_ERR;
    unsigned char value;

    if (pp_tp_i2c_chip_is_initialized(i2cchip) && pp_tp_scan_sensdev_do_scan(&pcfs->base)) {
		if (PP_SUC == (ret = pp_tp_i2c_chip_pre_com(i2cchip))) {
	        ret = pp_tp_i2c_chip_rx_byte(i2cchip, &value);
 	       pp_tp_i2c_chip_post_com(i2cchip);

	        if (ret == PP_SUC) {
			    if( cfg->config.all_raw_return == 1) {
			    	ret = value;
			    } else if( cfg->config.bit_raw_return == 1) {
			    	ret = (value & (1 << cfg->config.bit_shift)) ? 1 : 0;
			    } else {
			    // check discrete bit on the value returned  (0: plug-in, 1: plug-out)
			    	ret = ( (value & (1 << cfg->config.bit_shift))
			    			== (cfg->config.bit_in_value << cfg->config.bit_shift) ) ? 0 : 1;
			    };
	        } else {
	        	errno = EIO;
	        };
	    } else {
	    	errno = EIO;
	    };
    } else {
	};
    pp_tp_scan_sensdev_update_reading(&pcfs->base, ret);	
}

static void pcf8574_destroy(pcf8574_sensor_t* sensor) {
    /* unregister from scanning  */
    pp_tp_scan_sensdev_unregister(&sensor->base);
    
    pp_tp_i2c_chip_release(sensor->i2cchip);
    pp_tp_scan_sensdev_cleanup(&sensor->base);
    free(sensor);
}
