/**
 * w83627dhg_1.c
 *
 * Implements the Winbond83627DHG Monitor / Fan controller chip
 * Includes the i2c chip object as well as the sensor object
 * "P5BV-M", P5BP-E/4L", "RS100-E5/PI2"
 * (c) 2008 ASUS Tek, 2008/01/16, east_zhao@asus.com.tw
 */

/*
   P5BV-C/4L  P5BV-C   P5BP-E/4L  P5BV-E/SAS  P5BV-E     P5BV-M     P5GC-MR   P5BV       P5BV-SAS 
R1 56K OHM    56K OHM  15K OHM    15K OHM     15K OHM    15K OHM    56K OHM   15K OHM    15K OHM 
R2 10K OHM    10K OHM  2.49K OHM  2.49K OHM   2.49K OHM  2.49K OHM  10K OHM   2.49K OHM  2.49K OHM 
R3 22K OHM    22K OHM  10K OHM    10K OHM     10K OHM    10K OHM    22K OHM   10K OHM    10K OHM 
R4 10K OHM    10K OHM  3.9K OHM   4.99K OHM   4.99K OHM  3.9K OHM   10K OHM   4.99K OHM  4.99K OHM 
*/

#include <malloc.h>

#include <pp/base.h>
#include <pp/bmc/debug.h>
#include <pp/bmc/ipmi_sdr.h>
#include <pp/bmc/topo_factory.h>
#include <pp/bmc/tp_i2c_chip.h>
#include <pp/bmc/tp_scan_sensdev.h>
#include <pp/bmc/tp_pwm_act.h>
#include <pp/bmc/tp_gpio_dev.h>

#include "drivers/w83627dhg_1.h"

#define W83627DHG_PATCH_RECOVERY 0
#define W83627DHG_PATCH	1

/*
 * The chip, may contain global initialization code
 ****************************************************/

typedef struct pp_tp_i2c_chip_t {
    pp_tp_i2c_chip_t base;
} w83627dhg_1_chip_t;

static void w83627dhg_1_dtor(pp_tp_obj_t* o);

pp_tp_obj_t* pp_sensor_w83627dhg_1_ctor(const char* id, vector_t* args)
{
    pp_strstream_t err = PP_STRSTREAM_INITIALIZER;
    w83627dhg_1_chip_t* o = NULL;
    pp_tp_i2c_comdev_t* i2cdev;
    int i2caddr;
    pp_tp_cond_t* icond;

    if (pp_tp_arg_scanf(args, 0, &err, "o<i>do<sc>", &i2cdev, &i2caddr, &icond) != 3) {
	pp_bmc_log_perror("%s(): '%s' failed: %s", ___F, id, pp_strstream_buf(&err));
    } else {
    	o = malloc(sizeof(w83627dhg_1_chip_t));
    	pp_tp_i2c_chip_init(&o->base, PP_TP_I2C_CHIP, id, w83627dhg_1_dtor,
			    W83627DHG_MODEL_STRING, i2cdev, i2caddr, icond,   
			    NULL, NULL);
    }
    
    pp_strstream_free(&err);
    return (pp_tp_obj_t*)o;
}

static void w83627dhg_1_dtor(pp_tp_obj_t* o)
{
    pp_tp_i2c_chip_t* this = (pp_tp_i2c_chip_t*)o;
    assert(this != NULL);
    pp_tp_i2c_chip_cleanup(this);
    free(this);
}

/*
 * sensors
 ************************/
typedef struct w83627dhg_1_sens_type_cfg_s
{
    unsigned char sens_type;
    unsigned char reading_type;
    unsigned short analog_type;
    unsigned short unit_type;
    char sdr_rexp;
    void (*update)(pp_sensor_scannable_t*);
} w83627dhg_1_sens_type_cfg_t;

typedef struct w83627dhg_1_sens_cfg_s
{
    unsigned char reg;
    unsigned char bank;
    w83627dhg_1_sens_type_cfg_t* type;
    unsigned short sdr_mtol;
    unsigned short sdr_bacc;
    unsigned short sdr_bexp;    
    unsigned short sdr_rexp;
    unsigned short retry_cnt;

} w83627dhg_1_sens_cfg_t;

typedef struct w83627dhg_1_sens_s
{
    pp_tp_scan_sensdev_t base;
    pp_tp_i2c_chip_t* chip;
    w83627dhg_1_sens_cfg_t* cfg;
} w83627dhg_1_sens_t;

static void w83627dhg_1_sens_destroy(pp_tp_obj_t* o);
static ipmi_sdr_header_t* w83627dhg_1_sens_default_sdr(pp_tp_sensdev_t* s);
static void w83627dhg_1_update_temp_reading(pp_sensor_scannable_t*);
static void w83627dhg_1_update_tach_reading(pp_sensor_scannable_t*);
static void w83627dhg_1_update_voltage_reading(pp_sensor_scannable_t*);

w83627dhg_1_sens_type_cfg_t w83627dhg_1_sens_type_cfg_temp = {
    .sens_type    = IPMI_SENSOR_TYPE_TEMPERATURE,
    .reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
    .analog_type  = IPMI_ANALOG_DATA_FORMAT_UNSIGNED,
    .unit_type    = IPMI_UNIT_TYPE_DEGREES_C,
    .update       = w83627dhg_1_update_temp_reading,
};

w83627dhg_1_sens_type_cfg_t w83627dhg_1_sens_type_cfg_tacho = {
    .sens_type    = IPMI_SENSOR_TYPE_FAN,
    .reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
    .analog_type  = IPMI_ANALOG_DATA_FORMAT_UNSIGNED,
    .unit_type    = IPMI_UNIT_TYPE_RPM,
    .update       = w83627dhg_1_update_tach_reading,
};

w83627dhg_1_sens_type_cfg_t w83627dhg_1_sens_type_cfg_voltage = {
    .sens_type    = IPMI_SENSOR_TYPE_VOLTAGE,
    .reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
    .analog_type  = IPMI_ANALOG_DATA_FORMAT_UNSIGNED,
    .unit_type    = IPMI_UNIT_TYPE_VOLTS,
    .update       = w83627dhg_1_update_voltage_reading,
};

w83627dhg_1_sens_cfg_t w83627dhg_1_sens_cfg[] = {
    {
	/* 0: CPU Vcore */
	.reg = W83627DHG_REG_VALUE_VCOREA,
	.bank = 0,
	.type = &w83627dhg_1_sens_type_cfg_voltage,
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(8, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, 
	.sdr_rexp = -3,
        .retry_cnt = 0,
    },
    {
	/* 1: +12V */
	.reg = W83627DHG_REG_VALUE_VIN0,
	.bank = 0,
	.type = &w83627dhg_1_sens_type_cfg_voltage,
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(56, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, 
	.sdr_rexp = -3,
        .retry_cnt = 0,
    },
    {
	/* 2: AVCC */
	.reg = W83627DHG_REG_VALUE_AVCC,
	.bank = 0,
	.type = &w83627dhg_1_sens_type_cfg_voltage,
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(16, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, 
	.sdr_rexp = -3,
        .retry_cnt = 0,
    },
    {
	/* 3: +3V */
	.reg = W83627DHG_REG_VALUE_3VCC,
	.bank = 0,
	.type = &w83627dhg_1_sens_type_cfg_voltage,
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(16, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, 
	.sdr_rexp = -3,
        .retry_cnt = 0,
    },
    {
	/* 4:  VIN1 */
	.reg = W83627DHG_REG_VALUE_VIN1,
	.bank = 0,
	.type = &w83627dhg_1_sens_type_cfg_voltage,
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(16, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, 
	.sdr_rexp = -3,
        .retry_cnt = 0,
    },
    {
	/* 5: +5V */
	.reg = W83627DHG_REG_VALUE_VIN2,
	.bank = 0,
	.type = &w83627dhg_1_sens_type_cfg_voltage,
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(28, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, 
	.sdr_rexp = -3,
        .retry_cnt = 0,
    },
    {
	/* 6: VIN3 */
	.reg = W83627DHG_REG_VALUE_VIN3,
	.bank = 0,
	.type = &w83627dhg_1_sens_type_cfg_voltage,
	.sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(0x10, 1),
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(16, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, 
	.sdr_rexp = -3,
        .retry_cnt = 0,
    },
    {
	/* 7: 3VSB  */
	.reg = W83627DHG_REG_VALUE_5VSB,
	.bank = 5,
	.type = &w83627dhg_1_sens_type_cfg_voltage,
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(16, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, 
	.sdr_rexp = -3,
        .retry_cnt = 0,
    },
    {
	/* 8: VBAT */
	.reg = W83627DHG_REG_VALUE_VBAT,
	.bank = 5,
	.type = &w83627dhg_1_sens_type_cfg_voltage,
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(16, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, 
	.sdr_rexp = -3,
        .retry_cnt = 0,
    },
    {
	/* 9: REAR_FAN1 */
    	.reg = W83627DHG_REG_VALUE_SYSFANIN,
    	.bank = 0,
	.type = &w83627dhg_1_sens_type_cfg_tacho,
	.sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(80, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0,
	.sdr_rexp = 0,
        .retry_cnt = 0,
    },
    {
	/* 10: CPU_FAN1 */
    	.reg = W83627DHG_REG_VALUE_CPUFANIN0,
    	.bank = 0,
	.type = &w83627dhg_1_sens_type_cfg_tacho,
	.sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(80, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0,
	.sdr_rexp = 0,
        .retry_cnt = 0,
    },
    {
	/* 11: FRNT_FAN3  */
    	.reg = W83627DHG_REG_VALUE_CPUFANIN1,
    	.bank = 0,
	.type = &w83627dhg_1_sens_type_cfg_tacho,
	.sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(80, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0,
	.sdr_rexp = 0,
        .retry_cnt = 0,
    },
    {
	/* 12: FRNT_FAN1 */
    	.reg = W83627DHG_REG_VALUE_AUXFANIN0,
    	.bank = 0,
	.type = &w83627dhg_1_sens_type_cfg_tacho,
	.sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(80, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0,
	.sdr_rexp = 0,
        .retry_cnt = 0,
    },
    {
	/* 13: FRNT_FAN2 */
	.reg = W83627DHG_REG_VALUE_AUXFANIN1,
	.bank = 5,
	.type = &w83627dhg_1_sens_type_cfg_tacho,
	.sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(80, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0,
	.sdr_rexp = 0,
        .retry_cnt = 0,
    },
    {
	/* 14: MB Temperature */
    	.reg = W83627DHG_REG_VALUE_SYSTIN,
    	.bank = 0,
	.type = &w83627dhg_1_sens_type_cfg_temp,
	.sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(1, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0,
	.sdr_rexp = 0,
        .retry_cnt = 0,
    },
    {
	/* 15: CPU Temperature */
    	.reg = W83627DHG_REG_VALUE_CPUTIN_H,
    	.bank = 1,
      .type = &w83627dhg_1_sens_type_cfg_temp,
      .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(1, 0),
      .sdr_bacc	= PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
      .sdr_bexp = 0,
      .sdr_rexp = 0,
        .retry_cnt = 0,
    },
    {
	/* 16: AUXTIN */
    	.reg = W83627DHG_REG_VALUE_AUXTIN_H,
    	.bank = 2,
      .type = &w83627dhg_1_sens_type_cfg_temp,
      .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(1, 0),
      .sdr_bacc	= PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
      .sdr_bexp = 0,
      .sdr_rexp = 0,
        .retry_cnt = 0,
    },
};	
      
pp_tp_obj_t* pp_sensor_w83627dhg_1_sens_ctor(const char* id, vector_t* args)
{
    const char* errmsg = "%s(): '%s' failed: %s";
    w83627dhg_1_sens_t* this = NULL;
    pp_strstream_t err = PP_STRSTREAM_INITIALIZER;
    pp_tp_i2c_chip_t* chip;
    unsigned int sens_no;

    if (pp_tp_arg_scanf(args, 0, &err, "o<h>d", &chip, &sens_no) < 2) {
    	pp_bmc_log_perror(errmsg, ___F, id, pp_strstream_buf(&err));
    } else if (strcmp(chip->model, W83627DHG_MODEL_STRING)) {
	errno = EINVAL;
	pp_bmc_log_perror(errmsg, ___F, id, "incorrect chip model");
    }
    else if (sens_no > 16)
    {
    	errno = EINVAL;
    	pp_bmc_log_perror(errmsg, ___F, id, "invalid sensor number");
    }
    else
    {
    	w83627dhg_1_sens_cfg_t* cfg = &w83627dhg_1_sens_cfg[sens_no];
    	this = malloc(sizeof(w83627dhg_1_sens_t));
    	pp_tp_scan_sensdev_init(&this->base, PP_TP_SCAN_SENS_DEV, id,
				(void(*)(pp_tp_obj_t*))w83627dhg_1_sens_destroy,
				w83627dhg_1_sens_default_sdr, cfg->type->update,
				chip->init_cond);
    	this->chip = pp_tp_i2c_chip_duplicate(chip);
    	this->cfg = cfg;
    	
    	/* register for scanning  */
    	pp_tp_scan_sensdev_register(&this->base, 0);
    }
    pp_strstream_free(&err);
    return (pp_tp_obj_t*)this;
}

static void w83627dhg_1_sens_destroy(pp_tp_obj_t* o)
{
    w83627dhg_1_sens_t* this = (w83627dhg_1_sens_t*)o;
    assert(this != NULL);
    pp_tp_scan_sensdev_unregister(&this->base);
    pp_tp_i2c_chip_release(this->chip);
    pp_tp_scan_sensdev_cleanup(&this->base);
    free(this);
}

static ipmi_sdr_header_t* w83627dhg_1_sens_default_sdr(pp_tp_sensdev_t* s)
{
    w83627dhg_1_sens_t* this = (w83627dhg_1_sens_t*)s;
    w83627dhg_1_sens_cfg_t* cfg = this->cfg;
    ipmi_sdr_full_sensor_t* sdr;

    sdr = ipmi_sdr_full_part_create(cfg->type->sens_type,
				    cfg->type->reading_type,
				    cfg->type->unit_type,
				    cfg->type->analog_type,
				    cfg->sdr_mtol, 
				    cfg->sdr_bacc,
				    cfg->sdr_bexp, cfg->sdr_rexp,
				    ipmi_sdr_max_by_form(cfg->type->analog_type),
				    ipmi_sdr_min_by_form(cfg->type->analog_type));
    return (ipmi_sdr_header_t*)sdr;
}

static void w83627dhg_1_update_temp_reading(pp_sensor_scannable_t* s)
{
    w83627dhg_1_sens_t* this = PP_TP_SCANNABLE_2_OBJ_CAST(s, w83627dhg_1_sens_t);
    pp_tp_i2c_chip_t* chip = this->chip;
    w83627dhg_1_sens_cfg_t* cfg = this->cfg;
    int ret = -2;
    unsigned char value;
#if W83627DHG_PATCH_RECOVERY
    int i;
#endif
    
    if (pp_tp_i2c_chip_is_initialized(chip) && pp_tp_scan_sensdev_do_scan(&this->base))	
    {
	if (PP_SUC == (ret = pp_tp_i2c_chip_pre_com(chip))) 
        {
	    if(pp_tp_i2c_chip_tx_byte_data(chip, W83627DHG_REG_BANK_SEL, cfg->bank) == PP_SUC)
		if(pp_tp_i2c_chip_rx_byte_data(chip, cfg->reg, &value) == PP_SUC)
		    ret = value;

	#if W83627DHG_PATCH_RECOVERY
	    for(i=0; i<5; i++)
	    {
	       if ( (cfg->reg==W83627DHG_REG_VALUE_SYSTIN || cfg->reg==W83627DHG_REG_VALUE_CPUTIN_H) && 
	            ret > 120 )
	       {
		  pp_tp_i2c_chip_tx_byte_data(chip, W83627DHG_REG_BANK_SEL, 0);
	   	  pp_tp_i2c_chip_tx_byte_data(chip, W83627DHG_REG_FAN_DEVISOR_5, 0x04);

		  if (pp_tp_i2c_chip_tx_byte_data(chip, W83627DHG_REG_BANK_SEL, cfg->bank) == PP_SUC)
		     if(pp_tp_i2c_chip_rx_byte_data(chip, cfg->reg, &value) == PP_SUC)
			     ret = value;  
	       }
	    }
	    // avoid sensor read fail
	    if ( (i==5) && (cfg->reg==W83627DHG_REG_VALUE_SYSTIN || cfg->reg==W83627DHG_REG_VALUE_CPUTIN_H) && 
	            (ret > 120) )
	       ret = -2;
	#endif

	    if ( ret == 0 )
	       ret = -2;

	    // avoid power on detect fail
	    if ( ret >= 0x80 )	
            {	
	       ret = -2;
	    }
	    else if ( (cfg->reg==W83627DHG_REG_VALUE_SYSTIN ) && (ret >= 56) )	
            {	
		cfg->retry_cnt++;
		if ( cfg->retry_cnt == 0)
		   cfg->retry_cnt = 10;
	    }
	    else if ( (cfg->reg==W83627DHG_REG_VALUE_CPUTIN_H ) && (ret >= 0x50) )	
            {	
		cfg->retry_cnt++;
		if ( cfg->retry_cnt == 0)
		   cfg->retry_cnt = 10;
	    }

	    if (  (cfg->retry_cnt>0 && cfg->retry_cnt<=2) )		// fixed bug for divisor fail
	       ret = -2; 

	    pp_tp_i2c_chip_post_com(chip);
	}
	else
	   ret = -2;
    }

    pp_tp_scan_sensdev_update_reading(&this->base, ret);
}

static void w83627dhg_1_update_tach_reading(pp_sensor_scannable_t* s) {
    w83627dhg_1_sens_t* this = PP_TP_SCANNABLE_2_OBJ_CAST(s, w83627dhg_1_sens_t);
    pp_tp_i2c_chip_t* chip = this->chip;
    w83627dhg_1_sens_cfg_t* cfg = this->cfg;
    int ret = -2;
   
#if W83627DHG_PATCH_RECOVERY
    int i;
#endif
    unsigned char tachDivisor = 0;
    unsigned char value;
    unsigned char divTbl[] = {1,2,4,8,16,32,64,128};
    unsigned char fail_ret=0;
    unsigned char i2c_fail=1;


    if (pp_tp_i2c_chip_is_initialized(chip) && pp_tp_scan_sensdev_do_scan(&this->base)) {
    	if (PP_SUC == (ret = pp_tp_i2c_chip_pre_com(chip))) {
	    if(pp_tp_i2c_chip_tx_byte_data(chip, W83627DHG_REG_BANK_SEL, W83627DHG_REG_FAN_DEVISOR_BANK) != PP_SUC)
		goto finally;

	    switch(cfg->reg) {
	      case W83627DHG_REG_VALUE_SYSFANIN:
		  if(pp_tp_i2c_chip_rx_byte_data(chip, W83627DHG_REG_FAN_DEVISOR_1, &value) != PP_SUC)		
		      goto finally;

		  tachDivisor = (value >> 4) & 0x03;
		  if(pp_tp_i2c_chip_rx_byte_data(chip, W83627DHG_REG_FAN_DEVISOR_5, &value) != PP_SUC)
		      goto finally;

		#if W83627DHG_PATCH_RECOVERY
		  for(i=0; i<5; i++)
		  {
		     if ( (value >> 3 ) & 0x04 )
		     {
		        pp_tp_i2c_chip_tx_byte_data(chip, W83627DHG_REG_FAN_DEVISOR_5, 0x04);
		        pp_tp_i2c_chip_rx_byte_data(chip, W83627DHG_REG_FAN_DEVISOR_5, &value);
			fail_ret = 1;
		     }
		  }
		#endif
		  tachDivisor |= ((value >> 3) & 0x04);
		  value = tachDivisor & 0x07;
		  tachDivisor = divTbl[value];
		  break;
	      case W83627DHG_REG_VALUE_CPUFANIN0:
		  if(pp_tp_i2c_chip_rx_byte_data(chip, W83627DHG_REG_FAN_DEVISOR_1, &value) != PP_SUC)
		      goto finally;

		  tachDivisor = (value >> 6) & 0x03;
		  if(pp_tp_i2c_chip_rx_byte_data(chip, W83627DHG_REG_FAN_DEVISOR_5, &value) != PP_SUC)
		      goto finally;
		#if W83627DHG_PATCH_RECOVERY
		  for(i=0; i<5; i++)
		  {
		     if ( (value >> 4 ) & 0x04 )
		     {
		        pp_tp_i2c_chip_tx_byte_data(chip, W83627DHG_REG_FAN_DEVISOR_5, 0x04);
		        pp_tp_i2c_chip_rx_byte_data(chip, W83627DHG_REG_FAN_DEVISOR_5, &value);
			fail_ret = 1;
		     }
		  }
		#endif

		  tachDivisor |= ((value >> 4) & 0x04);
		  value = tachDivisor & 0x07;
		  tachDivisor = divTbl[value];
		  break;
	      case W83627DHG_REG_VALUE_AUXFANIN0:
		  if(pp_tp_i2c_chip_rx_byte_data(chip, W83627DHG_REG_FAN_DEVISOR_2, &value) != PP_SUC)
		      goto finally;
		  tachDivisor = (value >> 6) & 0x03;
		  if(pp_tp_i2c_chip_rx_byte_data(chip, W83627DHG_REG_FAN_DEVISOR_5, &value) != PP_SUC)
		      goto finally;
		#if W83627DHG_PATCH_RECOVERY
		  for(i=0; i<5; i++)
		  {		  
		     if ( (value >> 5 ) & 0x04 )
		     {
		        pp_tp_i2c_chip_tx_byte_data(chip, W83627DHG_REG_FAN_DEVISOR_5, 0x04);
		        pp_tp_i2c_chip_rx_byte_data(chip, W83627DHG_REG_FAN_DEVISOR_5, &value);
			fail_ret = 1;
		     }
		  }
		#endif
		  tachDivisor |= ((value >> 5) & 0x04);
		  value = tachDivisor & 0x07;
		  tachDivisor = divTbl[value];
		  break;
	      case W83627DHG_REG_VALUE_CPUFANIN1:
		  if(pp_tp_i2c_chip_rx_byte_data(chip, W83627DHG_REG_FAN_DEVISOR_4, &value) != PP_SUC)
		      goto finally;
		  tachDivisor = (value) & 0x03;
		  if(pp_tp_i2c_chip_rx_byte_data(chip, W83627DHG_REG_FAN_DEVISOR_3, &value) != PP_SUC)
		      goto finally;
		  tachDivisor |= ((value >> 5) & 0x04);
		  value = tachDivisor & 0x07;
		  tachDivisor = divTbl[value];
		  break;
	      case W83627DHG_REG_VALUE_AUXFANIN1:
		  if(pp_tp_i2c_chip_rx_byte_data(chip, W83627DHG_REG_FAN_DEVISOR_4, &value) != PP_SUC)
		      goto finally;
		  tachDivisor = (value >> 2) & 0x03;
		  tachDivisor |= ((value >> 5) & 0x04);
		  value = tachDivisor & 0x07;
		  tachDivisor = divTbl[value];
		  break;
	      default:
		  goto finally;
		  break;
	    }
  
	    if(pp_tp_i2c_chip_tx_byte_data(chip, W83627DHG_REG_BANK_SEL, cfg->bank) != PP_SUC)
		goto finally;

	    if (pp_tp_i2c_chip_rx_byte_data(chip, cfg->reg, &value) != PP_SUC)
		goto finally;
	    
	    i2c_fail = 0;
	    if( value == 0xFF || value == 0x00 )	// fixed bug for divisor fail
	    {
		cfg->retry_cnt++;
		if ( cfg->retry_cnt == 0)
		   cfg->retry_cnt = 10;
	    }
	    else
	    {
       		cfg->retry_cnt =0;
		ret = 1350000L / value / tachDivisor /80;

		if ( ret <= 0x0d || ret >=0xff )
                {
		   if ( ret <= 0x05 )
		      ret = -2 ;
		   else
		   {
		      cfg->retry_cnt++;
		      if ( cfg->retry_cnt == 0)
		         cfg->retry_cnt = 10;
		   }
		}            
	    }
	   
	    if ( value == 0 )
	    {
	       if ( fail_ret == 1 || (cfg->retry_cnt>0 && cfg->retry_cnt<=3) )		// fixed bug for divisor fail
	          ret = -2; 
	    }
	    else
	    {
	       if ( fail_ret == 1 || (cfg->retry_cnt>0 && cfg->retry_cnt<=5) )		// fixed bug for divisor fail
	          ret = -2; 
	    }
	    
	finally:
	    if ( i2c_fail == 1 )
	       ret = -2;
	    pp_tp_i2c_chip_post_com(chip);
	}
	else
	   ret = -2;
    } 
    pp_tp_scan_sensdev_update_reading(&this->base, ret);
}

static void w83627dhg_1_update_voltage_reading(pp_sensor_scannable_t* s) {
    w83627dhg_1_sens_t* this = PP_TP_SCANNABLE_2_OBJ_CAST(s, w83627dhg_1_sens_t);
    pp_tp_i2c_chip_t* chip = this->chip;
    w83627dhg_1_sens_cfg_t* cfg = this->cfg;
    int ret = -2;
    unsigned char value, value1;
    int fail_ret = 0;
    
    if (pp_tp_i2c_chip_is_initialized(chip) && pp_tp_scan_sensdev_do_scan(&this->base)) {
	if (PP_SUC == (ret = pp_tp_i2c_chip_pre_com(chip))) {

	#if W83627DHG_PATCH_RECOVERY
	      pp_tp_i2c_chip_tx_byte_data(chip, W83627DHG_REG_BANK_SEL, 0);
	      pp_tp_i2c_chip_rx_byte_data(chip, W83627DHG_REG_FAN_DEVISOR_5, &value);
	      if ( value != 0x04 )
	      {
	         pp_tp_i2c_chip_tx_byte_data(chip, W83627DHG_REG_FAN_DEVISOR_5, 0x04);
		 fail_ret = 1;
	      }
	#endif

	      if ( cfg->reg == W83627DHG_REG_VALUE_VBAT )
              {
	         pp_tp_i2c_chip_tx_byte_data(chip, W83627DHG_REG_BANK_SEL, 0);
	         pp_tp_i2c_chip_tx_byte_data(chip, W83627DHG_REG_FAN_DEVISOR_5, 0x05);
	         pp_tp_i2c_chip_rx_byte_data(chip, W83627DHG_REG_FAN_DEVISOR_5, &value);
	         pp_tp_i2c_chip_tx_byte_data(chip, W83627DHG_REG_FAN_DEVISOR_5, value | 0x01);	
	      }

	      if(pp_tp_i2c_chip_tx_byte_data(chip, W83627DHG_REG_BANK_SEL, cfg->bank) == PP_SUC)
		  if(pp_tp_i2c_chip_rx_byte_data(chip, cfg->reg, &value) == PP_SUC)
		      ret = value;

	#if W83627DHG_PATCH_RECOVERY
	      if ( cfg->reg == W83627DHG_REG_VALUE_VBAT )
              {
	         pp_tp_i2c_chip_tx_byte_data(chip, W83627DHG_REG_BANK_SEL, 0);
	         pp_tp_i2c_chip_tx_byte_data(chip, W83627DHG_REG_FAN_DEVISOR_5, 0x04);
	      }
	#endif

	      if ( cfg->reg == W83627DHG_REG_VALUE_VBAT )
              {
	         pp_tp_i2c_chip_tx_byte_data(chip, W83627DHG_REG_BANK_SEL, 0);
	         pp_tp_i2c_chip_tx_byte_data(chip, W83627DHG_REG_FAN_DEVISOR_5, 0x05);
	         pp_tp_i2c_chip_rx_byte_data(chip, W83627DHG_REG_FAN_DEVISOR_5, &value1);
	         pp_tp_i2c_chip_tx_byte_data(chip, W83627DHG_REG_FAN_DEVISOR_5, value1 & (~0x01));	
	      }
	  
	      if ( value == 0x00 || value == 0xff || ret == 0 )	// fixed bug for reset
	      {
		    fail_ret = 1;
	      }
	      else if ( cfg->reg == W83627DHG_REG_VALUE_VCOREA && ( value>=0xDE || value <=0x58 ) )	// Vcore
	      {
		    cfg->retry_cnt++;
		    if ( cfg->retry_cnt == 0)
		       cfg->retry_cnt = 10;
	      }
	      else if ( cfg->reg == W83627DHG_REG_VALUE_3VCC && ( value>=0xE2 || value <=0xAF) )	// 3.3V
	      {
		    cfg->retry_cnt++;
		    if ( cfg->retry_cnt == 0)
		       cfg->retry_cnt = 10;
	      }
	      else if ( cfg->reg == W83627DHG_REG_VALUE_VIN2 && ( value>=0xC5 || value <=0x97) )	// 5V
	      {
		    cfg->retry_cnt++;
		    if ( cfg->retry_cnt == 0)
		       cfg->retry_cnt = 10;
	      }
	      else if ( cfg->reg == W83627DHG_REG_VALUE_VIN0 && ( value>=0xE5 || value <=0xC1) )	// 12V
	      {
		    cfg->retry_cnt++;
		    if ( cfg->retry_cnt == 0)
		       cfg->retry_cnt = 10;
	      }
	      else if ( cfg->reg == W83627DHG_REG_VALUE_5VSB && ( value>=0xE2 || value <=0xAF) )	// 5VSB
	      {
		    cfg->retry_cnt++;
		    if ( cfg->retry_cnt == 0)
		       cfg->retry_cnt = 10;
	      }
	      else if ( cfg->reg == W83627DHG_REG_VALUE_VBAT && ( value>=0xCE || value <=0x9F ) )	// VBAT
	      {
		    cfg->retry_cnt++;
		    if ( cfg->retry_cnt == 0)
		       cfg->retry_cnt = 10;
	      }
	      else if ( value <=0x50 )	// fixed bug for reset
	      {
		    cfg->retry_cnt++;
		    if ( cfg->retry_cnt == 0)
		       cfg->retry_cnt = 10;
	      }
	      else
	      {
		 cfg->retry_cnt=0;
	      }

	      if ( fail_ret == 1 || (cfg->retry_cnt>0 && cfg->retry_cnt <= 3 ) )
		 ret = -2;

	   pp_tp_i2c_chip_post_com(chip);
	}
	else
	   ret = -2;
    } 
    pp_tp_scan_sensdev_update_reading(&this->base, ret);
}
